import os.path
import unittest

if __name__ == '__main__':
    start = os.path.dirname(os.path.dirname(__file__))
    suite = unittest.defaultTestLoader.discover(start)
    runner = unittest.TextTestRunner()
    runner.run(suite)


# coding=utf-8
import unittest
from mock import mock_open, patch
import exmldoc

sample_doc=u'''<?xml version="1.0" encoding="utf-8"?>
<exml-doc>
<schema>
<node name="topic">
</node>
</schema>
<body serialization="inline">
<text>
<sentence>
<topic>
<word form="This"/>
<word form="is"/>
<word form="a"/>
</topic>
<word form="sentence"/>
<word form="."/>
</sentence>
<topic>
<sentence>
<word form="No"/>
<word form="object"/>
<word form="has"/>
<word form="an"/>
<word form="ID"/>
<word form="."/>
</sentence>
<sentence>
<word form="Ümläüts"/>
<word form="and"/>
<word form="&#8220;non&#8221;-ISO"/>
<word form="(╯°□°）╯︵┻━┻"/>
<word form="work"/>
<word form="."/>
</sentence>
</topic>
</text>
</body>
</exml-doc>
'''.encode('utf-8')

sample_text_unicode = u'This is a sentence . No object has an ID . Ümläüts and “non”-ISO (╯°□°）╯︵┻━┻ work .'
sample_text_latin = u'This is a sentence . No object has an ID . Ümläüts and "non"-ISO (+°#°)+(+-+ work .'
sample_text_ascii = b'This is a sentence . No object has an ID . Umlauts and "non"-ISO (+deg#deg)+(+-+ work .'


class TestEXML(unittest.TestCase):
    def test_general(self):
        m = mock_open(read_data=sample_doc)
        with patch('exmldoc.open', m):
            doc = exmldoc.load('fake_data.exml.xml')
        m.assert_called_once_with('fake_data.exml.xml', 'rb')
        self.assertEqual(
            len(doc), 17,
            'length should be number of words')
        self.assertEqual(
            len(doc.get_objects_by_class(exmldoc.tree.Tree)), 3,
            'should read three sentences objects')
        self.assertEqual(
            len(doc.get_objects_by_level('sentence')), 3,
            'should read three sentence markables')
        self.assertEqual(
            len(doc.get_objects_by_level('text')), 1,
            'should read one text markable')
        self.assertEqual(
            len(doc.get_objects_by_level('topic')), 2,
            'should read two topic markables')

    def test_unicode(self):
        m = mock_open(read_data=sample_doc)
        with patch('exmldoc.open', m):
            doc = exmldoc.load('fake_data.exml.xml')
        self.assertEqual(
            doc.words,
            sample_text_unicode.split(),
            'unicode mode should produce unicode words')
        self.assertEqual(
            [term.word for term in doc.w_objs],
            sample_text_unicode.split(),
            'unicode mode should produce unicode term.words')

    def test_utf8(self):
        m = mock_open(read_data=sample_doc)
        with patch('exmldoc.open', m):
            doc = exmldoc.load('fake_data.exml.xml', encoding='utf-8')
        self.assertEqual(
            doc.words,
            sample_text_unicode.encode('utf-8').split(),
            'utf-8 mode should produce utf-8 word forms')
        self.assertEqual(
            [term.word for term in doc.w_objs],
            sample_text_unicode.encode('utf-8').split(),
            'utf-8 mode should produce utf-8 term.words')

    def test_latin1(self):
        m = mock_open(read_data=sample_doc)
        with patch('exmldoc.open', m):
            doc = exmldoc.load('fake_data.exml.xml', encoding='latin1')
        self.assertEqual(
            doc.words,
            sample_text_latin.encode('ISO-8859-15', 'ignore').split(),
            'latin1 mode should produce latin1 word forms')
        self.assertEqual(
            [term.word for term in doc.w_objs],
            sample_text_latin.encode('ISO-8859-15', 'ignore').split(),
            'latin1 mode should produce latin1 term.words')

    def test_ascii(self):
        m = mock_open(read_data=sample_doc)
        with patch('exmldoc.open', m):
            doc = exmldoc.load('fake_data.exml.xml', encoding='ascii')
        self.assertEqual(
            doc.words,
            sample_text_ascii.split(),
            'ascii mode should produce ascii word forms')
        self.assertEqual(
            [term.word for term in doc.w_objs],
            sample_text_ascii.split(),
            'ascii mode should produce ascii term.words')



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
'''
Creates a CQP-style column output from a EXML file.

Usage:
export2cqp [-f format] [-o output.txt] inputfile.exml.xml
'''
from __future__ import print_function

import sys
import os.path
import getopt
import glob

import exml


class ExportToCoNLL:
    """
    writes CoNLL-U (
    """

    def __init__(self, format='conllu', opts=None):
        self.format = format

        self.p_atts = []

    def write_conll(self, fname, f_out=None):
        if f_out is None:
            f_out = sys.stdout
        count = 0
        if self.format == 'conllu':
            print("# document: %s"%(fname,), file=f_out)
        for t in exml.read_trees_exml(fname):
            if self.format == 'conllu':
                if hasattr(t, 'sent_no'):
                    print("# sentence id: %s" % (t.sent_no,), file=f_out)
                elif hasattr(t, 'xml_id'):
                    print("# sentence id: %s" % (t.xml_id,), file=f_out)
                print("# sentence-text: %s"%(' '.join([n.word for n in t.terminals])))
            count += len(t.terminals)
            sent_start = t.terminals[0].span[0]
            for i, n in enumerate(t.terminals):
                print("%d\t%s\t%s\t%s\t%s" % (
                    i+1,
                    n.word, n.cat, n.lemma, n.morph), file=f_out)
            if self.want_deprels:
                for n in t.terminals:
                    lbl = getattr(n, 'syn_label', None)
                    if not hasattr(n, 'syn_parent') or n.syn_parent is None:
                        attach = 'ROOT'
                        lbl = 'ROOT'
                    else:
                        attach = '%+d' % (n.syn_parent.start - n.start)
                    if not hasattr(n, 'lemma') or n.lemma is None:
                        lemma = '_'
                    else:
                        lemma = n.lemma
                    print("%s\t%s\t%s\t%s\t%s\t%s" % (
                        n.word, n.cat, lemma, n.morph, lbl, attach), file=f_out)
            else:
                for n in t.terminals:
                    print("%s\t%s\t%s\t%s" % (
                        n.word, n.cat, n.lemma, n.morph), file=f_out)
            print("</s>", file=f_out)
        return count


def usage():
    print(__doc__)


def main():
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], 'P:S:o:')
    except getopt.GetoptError:
        usage()
        sys.exit(1)
    if len(args) != 1:
        usage()
        sys.exit(1)
    f_out = None
    for k, v in opts:
        if k == '-o':
            f_out = file(v, 'w')
    app = ExportToCoNLL(opts)
    if os.path.isdir(args[0]):
        # TODO do something sensible
        offsets_fname = os.path.join(args[0], 'offsets.txt')
        all_exml_fnames = glob.glob(os.path.join(args[0], '*.exml.xml'))
        if os.path.exists(offsets_fname):
            # Case 1: offsets.txt already exists
            # iterate over the files in offsets.txt, making sure
            # (i) no files are missed, (ii) the offsets still match
            all_files = set(all_exml_fnames)
            count_total = 0
            old_fname = '(START)'
            for l in open(offsets_fname, 'rb'):
                line = l.strip().split()
                fname = fname0 = line[0]
                if not fname.endswith('.exml.xml'):
                    fname = fname + '.exml.xml'
                f_path = os.path.join(args[0], fname)
                all_files.remove(f_path)
                print("<doc id=%s>" % (fname0,), file=f_out)
                count_file = app.write_conll(f_path, f_out)
                print("</doc>", file=f_out)
                posn_offset = int(line[1])
                if posn_offset != count_total:
                    print("ERROR: %s at position %d, should be %d" % (
                        fname, count_total, posn_offset), file=sys.stderr)
                    print("before: %s" % (old_fname,), file=sys.stderr)
                count_total = posn_offset + count_file
                old_fname = fname
            if all_files:
                # TODO: append new files to offsets.txt?
                for fname in sorted(all_files):
                    print("ERROR: %s is missing in offsets.txt" % (
                        fname,), file=sys.stderr)
        else:
            # Case 2: create offsets.txt
            f_offsets = open(offsets_fname, 'wb')
            count_total = 0
            for f_path in all_exml_fnames:
                fname0 = os.path.basename(f_path)
                if fname0.endswith('.exml.xml'):
                    fname0 = fname0[:-9]
                print("<doc id=%s>" % (fname0,), file=f_out)
                count_file = app.write_conll(f_path, f_out)
                print("</doc>", file=f_out)
                print("%s\t%d" % (
                    fname0, count_total), file=f_offsets)
                count_total += count_file
    else:
        app.write_conll(args[0], f_out)

if __name__ == '__main__':
    main()



# Copyright (c) 2007 RADLogic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Provide various handy Python functions.

Running this script directly will execute the doctests.

Functions:
int2bin(i, n) -- Convert integer to binary string.
bin2int(bin_string) -- Convert binary string to integer.
reverse(input_string) -- Reverse a string.
transpose(matrix) -- Transpose a list of lists.
polygon_area(points_list) -- Calculate the area of an arbitrary polygon.
timestamp() -- Return string containing current time stamp.
pt2str(point) -- Return prettier string version of point tuple.
gcf(a, b) -- Return the greatest common factor of two numbers.
lcm(a, b) -- Return the least common multiple of two numbers.
permutations(input_list) -- Generate all permutations of a list of items.
reduce_fraction(fraction) -- Reduce fraction (num, denom) to simplest form.
quantile(l, p) -- Return p quantile of list l. E.g. p=0.25 for q1.
trim(l) -- Discard values in list more than 1.5*IQR outside IQR.
nice_units(value) -- Return value converted to human readable units.
uniquify(seq) -- Return sequence with duplicate items in sequence seq removed.
reverse_dict(d) -- Return the dictionary with the items as keys and vice-versa.
lsb(x, n) -- Return the n least significant bits of x.
gray_encode(i) -- Gray encode the given integer.
random_vec(bits, max_value=None) -- Return a random binary vector.
binary_range(bits) -- Return list of all possible binary numbers width=bits.
float_range([start], stop, [step]) -- Return range of floats.
find_common_fixes(s1, s2) -- Find common (prefix, suffix) of two strings.
is_rotated(seq1, seq2) -- Return true if the list is a rotation of other list.
getmodule(obj) -- Return the module that contains the object definition of obj.
                  (use inspect.getmodule instead, though)
get_args(argv) -- Store command-line args in a dictionary.

This module requires Python >= 2.2

"""
__author__ = 'Tim Wegener <twegener@radlogic.com.au>'
__date__ = '$Date: 2007/03/27 03:15:06 $'
__version__ = '$Revision: 0.45 $ + YV custom fixes'
__credits__ = """
              David Chandler, for polygon area algorithm.
               (http://www.davidchandler.com/AreaOfAGeneralPolygon.pdf)
              """

import re
import sys
import time
import random

def int2bin(i, n):
    """Convert decimal integer i to n-bit binary number (string).

    >>> int2bin(0, 8)
    '00000000'

    >>> int2bin(123, 8)
    '01111011'

    >>> int2bin(123L, 8)
    '01111011'

    >>> int2bin(15, 2)
    Traceback (most recent call last):
    ValueError: Value too large for given number of bits.

    """
    hex2bin = {'0': '0000', '1': '0001', '2': '0010', '3': '0011',
               '4': '0100', '5': '0101', '6': '0110', '7': '0111',
               '8': '1000', '9': '1001', 'a': '1010', 'b': '1011',
               'c': '1100', 'd': '1101', 'e': '1110', 'f': '1111'}
    # Convert to hex then map each hex digit to binary equivalent.
    result = ''.join([hex2bin[x] for x in hex(i).lower().replace('l', '')[2:]])

    # Shrink result to appropriate length.
    # Raise an error if the value is changed by the truncation.
    if '1' in result[:-n]:
        raise ValueError("Value too large for given number of bits.")
    result = result[-n:]
    # Zero-pad if length longer than mapped result.
    result = '0' * (n - len(result)) + result
    return result


def bin2int(bin_string):
    """Convert binary number string to decimal integer.

    Note: Python > v2 has int(bin_string, 2)

    >>> bin2int('1111')
    15

    >>> bin2int('0101')
    5

    """
    return int(bin_string, 2)


def reverse(input_string):
    """Reverse a string. Useful for strings of binary numbers.

    >>> reverse('abc')
    'cba'

    """
    return input_string[::-1]


def transpose(matrix):
    """Transpose a list of lists.

    >>> transpose([['a', 'b', 'c'], ['d', 'e', 'f'], ['g', 'h', 'i']])
    [['a', 'd', 'g'], ['b', 'e', 'h'], ['c', 'f', 'i']]

    >>> transpose([['a', 'b', 'c'], ['d', 'e', 'f']])
    [['a', 'd'], ['b', 'e'], ['c', 'f']]

    >>> transpose([['a', 'b'], ['d', 'e'], ['g', 'h']])
    [['a', 'd', 'g'], ['b', 'e', 'h']]

    """
    result = zip(*matrix)
    # Convert list of tuples to list of lists.
    # map is faster than a list comprehension since it is being used with
    # a built-in function as an argument.
    result = map(list, result)
    return result


def polygon_area(points_list, precision=100):
    """Calculate area of an arbitrary polygon using an algorithm from the web.

    Return the area of the polygon as a positive float. 

    Arguments:
    points_list -- list of point tuples [(x0, y0), (x1, y1), (x2, y2), ...]
                   (Unclosed polygons will be closed automatically.
    precision -- Internal arithmetic precision (integer arithmetic).

    >>> polygon_area([(0, 0), (0, 1), (1, 1), (1, 2), (2, 2), (2, 0), (0, 0)])
    3.0

    Credits:
    Area of a General Polygon by David Chandler
    http://www.davidchandler.com/AreaOfAGeneralPolygon.pdf

    """
    # Scale up co-ordinates and convert them to integers.
    for i in range(len(points_list)):
        points_list[i] = (int(points_list[i][0] * precision),
                          int(points_list[i][1] * precision))
    # Close polygon if not closed.
    if points_list[-1] != points_list[0]:
        points_list.append(points_list[0])
    # Calculate area.
    area = 0
    for i in range(len(points_list) - 1):
        (x_i, y_i) = points_list[i]
        (x_i_plus_1, y_i_plus_1) = points_list[i + 1]
        area = area + (x_i_plus_1 * y_i) - (y_i_plus_1 * x_i)
    area = abs(area / 2)
    # Unscale area.
    area = float(area) / (precision**2)
    return area


def timestamp():
    """Return string containing current time stamp.

    Note: In Python 2 onwards can use time.asctime() with no arguments.

    """

    return time.asctime()


def pt2str(point):
    """Return prettier string version of point tuple.

    >>> pt2str((1.8, 1.9))
    '(1.8, 1.9)'

    """
    return "(%s, %s)" % (str(point[0]), str(point[1]))


def gcf(a, b, epsilon=1e-16):
    """Return the greatest common factor of a and b, using Euclidean algorithm.

    Arguments:
    a, b -- two numbers
            If both numbers are integers return an integer result, 
            otherwise return a float result.
    epsilon -- floats less than this magnitude are considered to be zero
               (default: 1e-16)

    Examples:

    >>> gcf(12, 34)
    2

    >>> gcf(13.5, 4)
    0.5

    >>> gcf(-2, 4)
    2

    >>> gcf(5, 0)
    5

    By (a convenient) definition:
    >>> gcf(0, 0)
    0

    """
    result = max(a, b)
    remainder = min(a, b)
    while remainder and abs(remainder) > epsilon:
        new_remainder = result % remainder
        result = remainder
        remainder = new_remainder
    return abs(result)


def lcm(a, b, precision=None):
    """Return the least common multiple of a and b, using the gcf function.

    Arguments:
    a, b -- two numbers. If both are integers return an integer result, 
            otherwise a return a float result.
    precision -- scaling factor if a and/or b are floats.

    >>> lcm(21, 6)
    42

    >>> lcm(2.5, 3.5)
    17.5

    >>> str(lcm(1.5e-8, 2.5e-8, precision=1e9))
    '7.5e-08'

    By (an arbitary) definition:
    >>> lcm(0, 0)
    0

    """
    # Note: Dummy precision argument is for backwards compatibility.
    # Do the division first.
    # (See http://en.wikipedia.org/wiki/Least_common_multiple )
    denom = gcf(a, b)
    if denom == 0:
        result = 0
    else:
        result = a * (b / denom)
    return result


def permutations(input_list):
    """Return a list containing all permutations of the input list.

    Note: This is a recursive function.

    >>> perms = permutations(['a', 'b', 'c'])
    >>> perms.sort()
    >>> for perm in perms:
    ...     print perm
    ['a', 'b', 'c']
    ['a', 'c', 'b']
    ['b', 'a', 'c']
    ['b', 'c', 'a']
    ['c', 'a', 'b']
    ['c', 'b', 'a']

    """
    out_lists = []
    if len(input_list) > 1:
        # Extract first item in list.
        item = input_list[0]
        # Find all permutations of remainder of list. (Recursive call.)
        sub_lists = permutations(input_list[1:])
        # For every permutation of the sub list...
        for sub_list in sub_lists:
            # Insert the extracted first item at every position of the list.
            for i in range(len(input_list)):
                new_list = sub_list[:]
                new_list.insert(i, item)
                out_lists.append(new_list)
    else:
        # Termination condition: only one item in input list.
        out_lists = [input_list]
    return out_lists


def reduce_fraction(fraction):
    """Reduce fraction tuple to simplest form. fraction=(num, denom)

    >>> reduce_fraction((14, 7))
    (2, 1)

    >>> reduce_fraction((-2, 4))
    (-1, 2)

    >>> reduce_fraction((0, 4))
    (0, 1)

    >>> reduce_fraction((4, 0))
    (1, 0)

    """
    (numerator, denominator) = fraction
    common_factor = abs(gcf(numerator, denominator))
    result = (numerator / common_factor, denominator / common_factor)
    return result


def quantile(l, p):
    """Return p quantile of list l. E.g. p=0.25 for q1.

    See:
    http://rweb.stat.umn.edu/R/library/base/html/quantile.html

    """
    l_sort = l[:]
    l_sort.sort()
    n = len(l)
    r = 1 + ((n - 1) * p)
    i = int(r)
    f = r - i
    if i < n:
        result = (1 - f) * l_sort[i - 1] + f * l_sort[i]
    else:
        result = l_sort[i - 1]
    return result


def trim(l):
    """Discard values in list more than 1.5*IQR outside IQR.

    (IQR is inter-quartile-range)

    This function uses rad_util.quantile

    1.5*IQR -- mild outlier
    3*IQR -- extreme outlier

    See:
    http://wind.cc.whecn.edu/~pwildman/statnew/section_7_-_exploratory_data_analysis.htm

    """
    l_sort = l[:]
    l_sort.sort()
    # Calculate medianscore  (based on stats.py lmedianscore by Gary Strangman)
    if len(l_sort) % 2 == 0:
        # If even number of scores, average middle 2.
        index = int(len(l_sort) / 2)  # Integer division correct
        median = float(l_sort[index] + l_sort[index - 1]) / 2
    else:
        # int divsion gives mid value when count from 0
        index = int(len(l_sort) / 2)
        median = l_sort[index]
    # Calculate IQR.
    q1 = quantile(l_sort, 0.25)
    q3 = quantile(l_sort, 0.75)
    iqr = q3 - q1
    iqr_extra = iqr * 1.5

    def in_interval(x, i=iqr_extra, q1=q1, q3=q3):
        return (x >= q1 - i and x <= q3 + i)
    l_trimmed = [x for x in l_sort if in_interval(x)]
    return l_trimmed


def nice_units(value, dp=0, sigfigs=None, suffix='', space=' ',
               use_extra_prefixes=False, use_full_name=False, mode='si'):
    """Return value converted to human readable units eg milli, micro, etc.

    Arguments:
    value -- number in base units
    dp -- number of decimal places to display (rounded)
    sigfigs -- number of significant figures to display (rounded)
               This overrides dp if set.
    suffix -- optional unit suffix to append to unit multiplier
    space -- seperator between value and unit multiplier (default: ' ')
    use_extra_prefixes -- use hecto, deka, deci and centi as well if set.
                          (default: False)
    use_full_name -- use full name for multiplier symbol,
                     e.g. milli instead of m
                     (default: False)
    mode -- 'si' for SI prefixes, 'bin' for binary multipliers (1024, etc.)
            (Default: 'si')

    SI prefixes from:
    http://physics.nist.gov/cuu/Units/prefixes.html
    (Greek mu changed to u.)
    Binary prefixes based on:
    http://physics.nist.gov/cuu/Units/binary.html

    >>> nice_units(2e-11)
    '20 p'

    >>> nice_units(2e-11, space='')
    '20p'

    """
    si_prefixes = {1e24:  ('Y', 'yotta'),
                   1e21:  ('Z', 'zetta'),
                   1e18:  ('E', 'exa'),
                   1e15:  ('P', 'peta'),
                   1e12:  ('T', 'tera'),
                   1e9:   ('G', 'giga'),
                   1e6:   ('M', 'mega'),
                   1e3:   ('k', 'kilo'),
                   1e-3:  ('m', 'milli'),
                   1e-6:  ('u', 'micro'),
                   1e-9:  ('n', 'nano'),
                   1e-12: ('p', 'pico'),
                   1e-15: ('f', 'femto'),
                   1e-18: ('a', 'atto'),
                   1e-21: ('z', 'zepto'),
                   1e-24: ('y', 'yocto')
                   }
    if use_extra_prefixes:
        si_prefixes.update({1e2:  ('h', 'hecto'),
                            1e1:  ('da', 'deka'),
                            1e-1: ('d', 'deci'),
                            1e-2: ('c', 'centi')
                            })
    bin_prefixes = {2**10: ('K', 'kilo'),
                    2**20: ('M', 'mega'),
                    2**30: ('G', 'mega'),
                    2**40: ('T', 'tera'),
                    2**50: ('P', 'peta'),
                    2**60: ('E', 'exa')
                    }
    if mode == 'bin':
        prefixes = bin_prefixes
    else:
        prefixes = si_prefixes
    prefixes[1] = ('', '')  # Unity.
    # Determine appropriate multiplier.
    multipliers = prefixes.keys()
    multipliers.sort()
    mult = None
    for i in range(len(multipliers) - 1):
        lower_mult = multipliers[i]
        upper_mult = multipliers[i + 1]
        if lower_mult <= value < upper_mult:
            mult_i = i
            break
    if mult is None:
        if value < multipliers[0]:
            mult_i = 0
        elif value >= multipliers[-1]:
            mult_i = len(multipliers) - 1
    mult = multipliers[mult_i]
    # Convert value for this multiplier.
    new_value = value / mult
    # Deal with special case due to rounding.
    if sigfigs is None:
        if mult_i < (len(multipliers) - 1) and \
                round(new_value, dp) == \
                round((multipliers[mult_i + 1] / mult), dp):
            mult = multipliers[mult_i + 1]
            new_value = value / mult
    # Concatenate multiplier symbol.
    if use_full_name:
        label_type = 1
    else:
        label_type = 0
    # Round and truncate to appropriate precision.
    if sigfigs is None:
        str_value = eval('"%.' + str(dp) + 'f" % new_value', locals(), {})
    else:
        str_value = eval('"%.' + str(sigfigs) + 'g" % new_value', locals(), {})
    return str_value + space + prefixes[mult][label_type] + suffix


def uniquify(seq, preserve_order=False):
    """Return sequence with duplicate items in sequence seq removed.

    The code is based on usenet post by Tim Peters.

    This code is O(N) if the sequence items are hashable, O(N**2) if not.

    Peter Bengtsson has a blog post with an empirical comparison of other
    approaches:
    http://www.peterbe.com/plog/uniqifiers-benchmark

    If order is not important and the sequence items are hashable then
    list(set(seq)) is readable and efficient.

    If order is important and the sequence items are hashable generator
    expressions can be used (in py >= 2.4) (useful for large sequences):
    seen = set()
    do_something(x for x in seq if x not in seen or seen.add(x))

    Arguments:
    seq -- sequence
    preserve_order -- if not set the order will be arbitrary
                      Using this option will incur a speed penalty.
                      (default: False)

    Example showing order preservation:

    >>> uniquify(['a', 'aa', 'b', 'b', 'ccc', 'ccc', 'd'], preserve_order=True)
    ['a', 'aa', 'b', 'ccc', 'd']

    Example using a sequence of un-hashable items:

    >>> uniquify([['z'], ['x'], ['y'], ['z']], preserve_order=True)
    [['z'], ['x'], ['y']]

    The sorted output or the non-order-preserving approach should equal
    that of the sorted order-preserving approach output:

    >>> unordered = uniquify([3, 3, 1, 2], preserve_order=False)
    >>> unordered.sort()
    >>> ordered = uniquify([3, 3, 1, 2], preserve_order=True)
    >>> ordered.sort()
    >>> ordered
    [1, 2, 3]
    >>> int(ordered == unordered)
    1

    """
    try:
        # Attempt fast algorithm.
        d = {}
        if preserve_order:
            # This is based on Dave Kirby's method (f8) noted in the post:
            # http://www.peterbe.com/plog/uniqifiers-benchmark
            return [x for x in seq if (x not in d) and not d.__setitem__(x, 0)]
        else:
            for x in seq:
                d[x] = 0
            return d.keys()
    except TypeError:
        # Have an unhashable object, so use slow algorithm.
        result = []
        app = result.append
        for x in seq:
            if x not in result:
                app(x)
        return result


# Alias to noun form for backward compatibility.
unique = uniquify


def reverse_dict(d):
    """Reverse a dictionary so the items become the keys and vice-versa.

    Note: The results will be arbitrary if the items are not unique.

    >>> d = reverse_dict({'a': 1, 'b': 2})
    >>> d_items = d.items()
    >>> d_items.sort()
    >>> d_items
    [(1, 'a'), (2, 'b')]

    """
    result = {}
    for key, value in d.items():
        result[value] = key
    return result


def lsb(x, n):
    """Return the n least significant bits of x.

    >>> lsb(13, 3)
    5

    """
    return x & ((2 ** n) - 1)


def gray_encode(i):
    """Gray encode the given integer."""

    return i ^ (i >> 1)


def random_vec(bits, max_value=None):
    """Generate a random binary vector of length bits and given max value."""

    vector = ""
    for _ in range(int(bits / 10) + 1):
        i = int((2**10) * random.random())
        vector += int2bin(i, 10)

    if max_value and (max_value < 2 ** bits - 1):
        vector = int2bin((int(vector, 2) / (2 ** bits - 1)) * max_value, bits)

    return vector[0:bits]


def binary_range(bits):
    """Return a list of all possible binary numbers in order with width=bits. 

    It would be nice to extend it to match the
    functionality of python's range() built-in function.

    """
    l = []
    v = ['0'] * bits

    toggle = [1] + [0] * bits

    while toggle[bits] != 1:
        v_copy = v[:]
        v_copy.reverse()
        l.append(''.join(v_copy))

        toggle = [1] + [0] * bits
        i = 0
        while i < bits and toggle[i] == 1:
            if toggle[i]:
                if v[i] == '0':
                    v[i] = '1'
                    toggle[i + 1] = 0
                else:
                    v[i] = '0'
                    toggle[i + 1] = 1
            i += 1
    return l


def float_range(start, stop=None, step=None):
    """Return a list containing an arithmetic progression of floats.

    Return a list of floats between 0.0 (or start) and stop with an
    increment of step. 

    This is in functionality to python's range() built-in function 
    but can accept float increments.

    As with range(), stop is omitted from the list.

    """
    if stop is None:
        stop = float(start)
        start = 0.0

    if step is None:
        step = 1.0

    cur = float(start)
    l = []
    while cur < stop:
        l.append(cur)
        cur += step

    return l


def find_common_fixes(s1, s2):
    """Find common (prefix, suffix) of two strings.

    >>> find_common_fixes('abc', 'def')
    ('', '')

    >>> find_common_fixes('abcelephantdef', 'abccowdef')
    ('abc', 'def')

    >>> find_common_fixes('abcelephantdef', 'abccow')
    ('abc', '')

    >>> find_common_fixes('elephantdef', 'abccowdef')
    ('', 'def')

    """
    prefix = []
    suffix = []

    i = 0
    common_len = min(len(s1), len(s2))
    while i < common_len:
        if s1[i] != s2[i]:
            break

        prefix.append(s1[i])
        i += 1

    i = 1
    while i < (common_len + 1):
        if s1[-i] != s2[-i]:
            break

        suffix.append(s1[-i])
        i += 1

    suffix.reverse()

    prefix = ''.join(prefix)
    suffix = ''.join(suffix)

    return (prefix, suffix)


def is_rotated(seq1, seq2):
    """Return true if the first sequence is a rotation of the second sequence.

    >>> seq1 = ['A', 'B', 'C', 'D']
    >>> seq2 = ['C', 'D', 'A', 'B']
    >>> int(is_rotated(seq1, seq2))
    1

    >>> seq2 = ['C', 'D', 'B', 'A']
    >>> int(is_rotated(seq1, seq2))
    0

    >>> seq1 = ['A', 'B', 'C', 'A']
    >>> seq2 = ['A', 'A', 'B', 'C']
    >>> int(is_rotated(seq1, seq2))
    1

    >>> seq2 = ['A', 'B', 'C', 'A']
    >>> int(is_rotated(seq1, seq2))
    1

    >>> seq2 = ['A', 'A', 'C', 'B']
    >>> int(is_rotated(seq1, seq2))
    0

    """
    # Do a sanity check.
    if len(seq1) != len(seq2):
        return False
    # Look for occurrences of second sequence head item in first sequence.
    start_indexes = []
    head_item = seq2[0]
    for index1 in range(len(seq1)):
        if seq1[index1] == head_item:
            start_indexes.append(index1)
    # Check that wrapped sequence matches.
    double_seq1 = seq1 + seq1
    for index1 in start_indexes:
        if double_seq1[index1:index1 + len(seq1)] == seq2:
            return True
    return False


def getmodule(obj):
    """Return the module that contains the object definition of obj.

    Note: Use inspect.getmodule instead.

    Arguments:
    obj -- python obj, generally a class or a function

    Examples:

    A function:
    >>> module = getmodule(random.choice)
    >>> module.__name__
    'random'
    >>> module is random
    1

    A class:
    >>> module = getmodule(random.Random)
    >>> module.__name__
    'random'
    >>> module is random
    1

    A class inheriting from a class in another module:
    (note: The inheriting class must define at least one function.)
    >>> class MyRandom(random.Random):
    ...     def play(self):
    ...         pass
    >>> module = getmodule(MyRandom)
    >>> if __name__ == '__main__':
    ...     name = 'rad_util'
    ... else:
    ...     name = module.__name__
    >>> name
    'rad_util'
    >>> module is sys.modules[__name__]
    1

    Discussion:
    This approach is slightly hackish, and won't work in various situations.
    However, this was the approach recommended by GvR, so it's as good as
    you'll get.

    See GvR's post in this thread:
    http://groups.google.com.au/group/comp.lang.python/browse_thread/thread/966a7bdee07e3b34/c3cab3f41ea84236?lnk=st&q=python+determine+class+module&rnum=4&hl=en#c3cab3f41ea84236

    """
    if hasattr(obj, 'func_globals'):
        func = obj
    else:
        # Handle classes.
        func = None
        for item in obj.__dict__.values():
            if hasattr(item, 'func_globals'):
                func = item
                break
        if func is None:
            raise ValueError("No functions attached to object: %r" % obj)
    module_name = func.__globals__['__name__']
    # Get module.
    module = sys.modules[module_name]
    return module


def round_grid(value, grid, mode=0):
    """Round off the given value to the given grid size.

    Arguments:
    value -- value to be roudne
    grid -- result must be a multiple of this
    mode -- 0 nearest, 1 up, -1 down

    Examples:

    >>> round_grid(7.5, 5)
    10

    >>> round_grid(7.5, 5, mode=-1)
    5

    >>> round_grid(7.3, 5, mode=1)
    10

    >>> round_grid(7.3, 5.0, mode=1)
    10.0

    """
    off_grid = value % grid
    if mode == 0:
        add_one = int(off_grid >= (grid / 2.0))
    elif mode == 1 and off_grid:
        add_one = 1
    elif mode == -1 and off_grid:
        add_one = 0
    result = ((int(value / grid) + add_one) * grid)
    return result


def get_args(argv):
    """Store command-line args in a dictionary.

    -, -- prefixes are removed
    Items not prefixed with - or -- are stored as a list, indexed by 'args'

    For options that take a value use --option=value

    Consider using optparse or getopt (in Python standard library) instead.

    """
    d = {}
    args = []

    for arg in argv:

        if arg.startswith('-'):
            parts = re.sub(r'^-+', '', arg).split('=')
            if len(parts) == 2:
                d[parts[0]] = parts[1]
            else:
                d[parts[0]] = None
        else:
            args.append(arg)

    d['args'] = args

    return d


if __name__ == '__main__':
    import doctest
    doctest.testmod(sys.modules['__main__'])



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
'''
Creates a CQP-style column output from a EXML file.

Usage:
export2cqp [-P att] [-S satt] [-o output.txt] inputfile.exml.xml
'''
from __future__ import print_function

import sys
import os.path
import getopt
import glob

import exmldoc

if sys.version_info[0] >= 3:
    xrange = range

class ExportToCQP:

    def __init__(self, opts=None):
        self.p_atts = []
        self.s_atts = []
        if opts is None:
            self.want_deprels = False
        else:
            # add p_atts and s_atts etc.
            for k, v in opts:
                if k == '-S':
                    self.s_atts.append(v)
                elif k == '-P':
                    self.p_atts.append(v)
            self.want_deprels = False

    def write_cqp(self, reader, f_out=None):
        if f_out is None:
            f_out = sys.stdout
        count = 0
        level_map = {
            'sentence': ('s', None)
        }
        for s_att in self.s_atts:
            level_map[s_att] = (s_att, None)
        for ev in reader.inline_events(level_map):
            if ev[0] == 'start':
                # start S-attribute
                tag = ev[1]
                mapped_tag, write_fn = level_map[tag]
                attrs = ev[2]
                parts = []
                if write_fn is not None:
                    write_fn(tag, attrs)
                else:
                    for k,v in attrs:
                        if k == 'xml:id':
                            k = 'id'
                            # try to spot auto-generated IDs
                            if (v[0] == 'm' and len(v) == 7 or
                                    v.startswith('__tmp_')):
                                continue
                        parts.append(' %s=%s'%(k, v))
                    print('<%s%s>'%(mapped_tag, ''.join(parts)), file=f_out)
            elif ev[0] == 'end':
                # end S-attribute
                tag = ev[1]
                mapped_tag, write_fn = level_map[tag]
                print('</%s>'%(mapped_tag), file=f_out)
            elif ev[0] == 'terminal':
                n = ev[1]
                if not hasattr(n, 'lemma') or n.lemma is None:
                    lemma = '_'
                else:
                    lemma = n.lemma
                columns = [n.word, n.cat, lemma, n.morph]
                if self.want_deprels:
                    lbl = getattr(n, 'syn_label', '_')
                    if not hasattr(n, 'syn_parent') or n.syn_parent is None:
                        attach = 'ROOT'
                        lbl = 'ROOT'
                    else:
                        attach = '%+d'%(n.syn_parent.span[0] - n.span[0])
                    columns += [lbl, attach]
                print('\t'.join([str(x) for x in columns]), file=f_out)
                count += 1
        return count


def usage():
    print(__doc__)

def skip_through(reader):
    """
    continues reading a document until the end of the document
    :param reader: a document reader
    :param doc:    the document in question
    :return: the number of tokens in the document
    """
    last_stop = 0
    while True:
        try:
            next_stop = reader.addNext()
        except StopIteration:
            return len(reader.doc.words)


def process_directory(dirname, create_doc=None):
    """
    processes a directory of EXML files, consuming or creating an offsets.txt file
    :param dirname:
    :param create_doc:
    :return: an iterable of (fname, doc, reader) tuples
    """
    if create_doc is None:
        create_doc = lambda: exmldoc.make_syntax_doc(want_deps=True)
    offsets_fname = os.path.join(dirname, 'offsets.txt')
    all_exml_fnames = set(glob.glob(os.path.join(dirname, '*.exml.xml')))
    count_total = 0
    count_file = 0
    reader = None
    old_fname = '(START)'
    if os.path.exists(offsets_fname):
        for l in open(offsets_fname):
            line = l.strip().split()
            fname = fname0 = line[0]
            if not fname.endswith('.exml.xml'):
                fname = fname + '.exml.xml'
            if count_file is not None:
                count_total += count_file
                posn_offset = int(line[1])
                if posn_offset != count_total:
                    print("ERROR: %s at position %d, should be %d" % (
                        fname, count_total, posn_offset), file=sys.stderr)
                    print("before: %s" % (old_fname,), file=sys.stderr)
            else:
                count_total = int(line[1])
            f_path = os.path.join(dirname, fname)
            all_exml_fnames.remove(f_path)
            doc = create_doc()
            reader = exmldoc.XMLCorpusReader(doc, f_path)
            yield fname0, doc, reader
            if reader.at_end:
                count_file = len(doc.w_objs)
            else:
                count_file = None
            old_fname = fname
    if all_exml_fnames:
        if count_file is None:
            count_file = skip_through(reader)
        count_total += count_file
        with open(offsets_fname, 'a') as f_offsets:
            for fname in sorted(all_exml_fnames):
                fname0 = os.path.basename(fname)
                f_path = os.path.join(dirname, fname)
                print("%s\t%s"%(fname0, count_total), file=f_offsets)
                doc = create_doc()
                reader = exmldoc.XMLCorpusReader(doc, f_path)
                yield fname, doc, reader
                if reader.at_end:
                    count_file = len(doc.w_objs)
                else:
                    count_file = skip_through(reader)
                count_total += count_file

def main():
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], 'P:S:o:')
    except getopt.GetoptError:
        usage()
        sys.exit(1)
    if len(args) != 1:
        usage()
        sys.exit(1)
    f_out = None
    for k, v in opts:
        if k == '-o':
            f_out = open(v, 'w')
    app = ExportToCQP(opts)
    if os.path.isdir(args[0]):
        for fname, doc, reader in process_directory(args[0]):
            print("<doc id=%s>" % (fname,), file=f_out)
            app.write_cqp(reader, f_out)
            print("</doc>", file=f_out)
    else:
        doc = exmldoc.make_syntax_doc(want_deps=True)
        reader = exmldoc.XMLCorpusReader(doc, args[0])
        app.write_cqp(reader, f_out)


def write_cqp_entry(args=None):
    if args is None:
        args = sys.argv[1:]
    opts, args = getopt.gnu_getopt(args, 'dc:')
    env = {'cqp_dir': '/export/local/yannick/cqp',
           'want_deprels': False}
    for k, v in opts:
        if k == '-d':
            env['want_deprels'] = True
        elif k == '-c':
            env['cqp_dir'] = v
    env['corpus'] = args[0].upper()
    env['corpus_lower'] = args[0].lower()
    if env['cqp_dir'][-1] == '/':
        env['cqp_dir'] = env['cqp_dir'][:-1]
    print('''##
## registry entry for corpus %(corpus_lower)s
##

# long descriptive name for the corpus
NAME "%(corpus_lower)s"
# corpus ID (must be lowercase in registry!)
ID   %(corpus_lower)s
# path to binary data files
HOME %(cqp_dir)s/%(corpus_lower)s
# optional info file (displayed by "info;" command in CQP)
INFO %(cqp_dir)s/%(corpus_lower)s/.info

# corpus properties provide additional information about the corpus:
##:: charset  = "latin1" # change if your corpus uses different charset
##:: language = "??"     # insert ISO code for language (de, en, fr, ...)


##
## p-attributes (token annotations)
##

ATTRIBUTE word
ATTRIBUTE pos
ATTRIBUTE lemma
ATTRIBUTE morph
''' % env)
    if env['want_deprels']:
        print('''ATTRIBUTE deprel
ATTRIBUTE attach

''')
    print('''##
## s-attributes (structural markup)
##

# <s id=".."> ... </s>
# (no recursive embedding allowed)
STRUCTURE s                    # [annotations]
STRUCTURE s_id                 # [annotations]

# <text id=".."> ... </text>
# (no recursive embedding allowed)
STRUCTURE text                 # [annotations]
STRUCTURE text_id              # [annotations]
''')


if __name__ == '__main__':
    main()



# topsort - dependency (topological) sorting and cycle finding functions
# Copyright (C) 2007 RADLogic
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# See http://www.fsf.org/licensing/licenses/lgpl.txt for full license text.
"""Provide toplogical sorting (i.e. dependency sorting) functions.

The topsort function is based on code posted on Usenet by Tim Peters.

Modifications:
- added doctests
- changed some bits to use current Python idioms
  (listcomp instead of filter, +=/-=, inherit from Exception)
- added a topsort_levels version that ports items in each dependency level
  into a sub-list
- added find_cycles to aid in cycle debugging

Run this module directly to run the doctests (unittests).
Make sure they all pass before checking in any modifications.

Requires Python >= 2.2
(For Python 2.2 also requires separate sets.py module)

This requires the rad_util.py module.

"""

# Provide support for Python 2.2*
from __future__ import generators
from __future__ import absolute_import

__version__ = '$Revision: 0.9 $'
__date__ = '$Date: 2007/03/27 04:15:26 $'
__credits__ = '''Tim Peters -- original topsort code
Tim Wegener -- doctesting, updating to current idioms, topsort_levels,
               find_cycles
'''

# Make Python 2.3 sets look like Python 2.4 sets.
try:
    set
except NameError:
    from sets import Set as set

from .rad_util import is_rotated


class CycleError(Exception):
    """Cycle Error"""
    pass


def topsort(pairlist):
    """Topologically sort a list of (parent, child) pairs.

    Return a list of the elements in dependency order (parent to child order).

    >>> print topsort( [(1,2), (3,4), (5,6), (1,3), (1,5), (1,6), (2,5)] ) 
    [1, 2, 3, 5, 4, 6]

    >>> print topsort( [(1,2), (1,3), (2,4), (3,4), (5,6), (4,5)] )
    [1, 2, 3, 4, 5, 6]

    >>> print topsort( [(1,2), (2,3), (3,2)] )
    Traceback (most recent call last):
    CycleError: ([1], {2: 1, 3: 1}, {2: [3], 3: [2]})

    """
    num_parents = {}  # element -> # of predecessors
    children = {}  # element -> list of successors
    for parent, child in pairlist:
        # Make sure every element is a key in num_parents.
        if parent not in num_parents:
            num_parents[parent] = 0
        if child not in num_parents:
            num_parents[child] = 0

        # Since child has a parent, increment child's num_parents count.
        num_parents[child] += 1

        # ... and parent gains a child.
        children.setdefault(parent, []).append(child)

    # Suck up everything without a parent.
    answer = [x for x in num_parents.keys() if num_parents[x] == 0]

    # For everything in answer, knock down the parent count on its children.
    # Note that answer grows *in* the loop.
    for parent in answer:
        del num_parents[parent]
        if parent in children:
            for child in children[parent]:
                num_parents[child] -= 1
                if num_parents[child] == 0:
                    answer.append(child)
            # Following "del" isn't needed; just makes
            # CycleError details easier to grasp.
            del children[parent]

    if num_parents:
        # Everything in num_parents has at least one child ->
        # there's a cycle.
        raise CycleError(answer, num_parents, children)
    return answer


def topsort_levels(pairlist):
    """Topologically sort a list of (parent, child) pairs into depth levels.

    This returns a generator. 
    Turn this into a an iterator using the iter built-in function.
    (if you iterate over the iterator, each element gets generated when
    it is asked for, rather than generating the whole list up-front.)

    Each generated element is a list of items at that dependency level.

    >>> dependency_pairs = [(1,2), (3,4), (5,6), (1,3), (1,5), (1,6), (2,5)]
    >>> for level in iter(topsort_levels( dependency_pairs )):
    ...    print level
    [1]
    [2, 3]
    [4, 5]
    [6]

    >>> dependency_pairs = [(1,2), (1,3), (2,4), (3,4), (5,6), (4,5)]
    >>> for level in iter(topsort_levels( dependency_pairs )):
    ...    print level
    [1]
    [2, 3]
    [4]
    [5]
    [6]

    >>> dependency_pairs = [(1,2), (2,3), (3,4), (4, 3)]
    >>> try:
    ...     for level in iter(topsort_levels( dependency_pairs )):
    ...         print level
    ... except CycleError, exc:
    ...     print 'CycleError:', exc
    [1]
    [2]
    CycleError: ({3: 1, 4: 1}, {3: [4], 4: [3]})


    The cycle error should look like.
    CycleError: ({3: 1, 4: 1}, {3: [4], 4: [3]})
    # todo: Make the doctest more robust (i.e. handle arbitrary dict order).

    """
    num_parents = {}  # element -> # of predecessors
    children = {}  # element -> list of successors
    for parent, child in pairlist:
        # Make sure every element is a key in num_parents.
        if parent not in num_parents:
            num_parents[parent] = 0
        if child not in num_parents:
            num_parents[child] = 0

        # Since child has a parent, increment child's num_parents count.
        num_parents[child] += 1

        # ... and parent gains a child.
        children.setdefault(parent, []).append(child)

    return topsort_levels_core(num_parents, children)


def topsort_levels_core(num_parents, children):
    """Topologically sort a bunch of interdependent items based on dependency.

    This returns a generator.
    Turn this into a an iterator using the iter built-in function.
    (if you iterate over the iterator, each element gets generated when
    it is asked for, rather than generating the whole list up-front.)

    Each generated element is a list of items at that dependency level.

    >>> list(topsort_levels_core(
    ...          {1: 0, 2: 1, 3: 1, 4: 1, 5: 2, 6: 2},
    ...          {1: [2, 3, 5, 6], 2: [5], 3: [4], 4: [], 5: [6]}))
    [[1], [2, 3], [4, 5], [6]]

    >>> list(topsort_levels_core(
    ...          {1: 0, 2: 2, 3: 1},
    ...          {1: [2], 2: [3], 3: [2]}))
    Traceback (most recent call last):
    CycleError: ({2: 1, 3: 1}, {2: [3], 3: [2]})

    This function has a more complicated interface than topsort_levels,
    but is useful if the data is easier to generate in this form.

    Arguments:
    num_parents -- key: item, value: number of parents (predecessors)
    children -- key: item, value: list of children (successors)

    """
    while 1:
        # Suck up everything without a predecessor.
        level_parents = [x for x in num_parents.keys() if num_parents[x] == 0]

        if not level_parents:
            break

        # Offer the next generated item,
        # which is a list of the items at this dependency level.
        yield level_parents

        # For everything item in this level,
        # decrement the parent count,
        # since we have accounted for its parent.
        for level_parent in level_parents:

            del num_parents[level_parent]

            if level_parent in children:
                for level_parent_child in children[level_parent]:
                    num_parents[level_parent_child] -= 1
                del children[level_parent]

    if num_parents:
        # Everything in num_parents has at least one child ->
        # there's a cycle.
        raise CycleError(num_parents, children)
    else:
        # This is the end of the generator.
        raise StopIteration


def find_cycles(parent_children):
    """Yield cycles. Each result is a list of items comprising a cycle.

    Use a 'stack' based approach to find all the cycles.
    This is a generator, so yields each cycle as it finds it.

    It is implicit that the last item in each cycle list is a parent of the
    first item (thereby forming a cycle).

    Arguments:
    parent_children -- parent -> collection of children

    Simplest cycle:
    >>> cycles = list(find_cycles({'A': ['B'], 'B': ['A']}))
    >>> len(cycles)
    1
    >>> cycle = cycles[0]
    >>> cycle.sort()
    >>> print cycle
    ['A', 'B']

    Simplest cycle with extra baggage at the start and the end:
    >>> cycles = list(find_cycles(parent_children={'A': ['B'],
    ...                                            'B': ['C'],
    ...                                            'C': ['B', 'D'],
    ...                                            'D': [],
    ...                                            }))
    >>> len(cycles)
    1
    >>> cycle = cycles[0]
    >>> cycle.sort()
    >>> print cycle
    ['B', 'C']

    Double cycle:
    >>> cycles = list(find_cycles(parent_children={'A': ['B'],
    ...                                            'B': ['C1', 'C2'],
    ...                                            'C1': ['D1'],
    ...                                            'D1': ['E1'],
    ...                                            'E1': ['D1'],
    ...                                            'C2': ['D2'],
    ...                                            'D2': ['E2'],
    ...                                            'E2': ['D2'],
    ...                                            }))
    >>> len(cycles)
    2
    >>> for cycle in cycles:
    ...     cycle.sort()
    >>> cycles.sort()
    >>> cycle1 = cycles[0]
    >>> cycle1.sort()
    >>> print cycle1
    ['D1', 'E1']
    >>> cycle2 = cycles[1]
    >>> cycle2.sort()
    >>> print cycle2
    ['D2', 'E2']

    Simple cycle with children not specified for one item:
    # todo: Should this barf instead?
    >>> cycles = list(find_cycles(parent_children={'A': ['B'],
    ...                                            'B': ['A'],
    ...                                            'C': ['D']}))
    >>> len(cycles)
    1
    >>> cycle = cycles[0]
    >>> cycle.sort()
    >>> print cycle
    ['A', 'B']

    Diamond cycle
    >>> cycles = list(find_cycles(parent_children={'A': ['B1', 'B2'],
    ...                                            'B1': ['C'],
    ...                                            'B2': ['C'],
    ...                                            'C': ['A', 'B1']}))
    >>> len(cycles)
    3
    >>> sorted_cycles = []
    >>> for cycle in cycles:
    ...     cycle = list(cycle)
    ...     cycle.sort()
    ...     sorted_cycles.append(cycle)
    >>> sorted_cycles.sort()
    >>> for cycle in sorted_cycles:
    ...     print cycle
    ['A', 'B1', 'C']
    ['A', 'B2', 'C']
    ['B1', 'C']

    Hairy case (order can matter if something is wrong):
    (Note order of B and C in the list.)
    >>> cycles = list(find_cycles(parent_children={
    ...                                           'TD': ['DD'],
    ...                                           'TC': ['DC'],
    ...                                           'DC': ['DQ'],
    ...                                           'C': ['DQ'],
    ...                                           'DQ': ['IA', 'TO'],
    ...                                           'IA': ['A'],
    ...                                           'A': ['B', 'C'],
    ...                                           }))
    >>> len(cycles)
    1
    >>> cycle = cycles[0]
    >>> cycle.sort()
    >>> print cycle
    ['A', 'C', 'DQ', 'IA']

    """
    cycles = []
    visited_nodes = set()

    for parent in parent_children:
        if parent in visited_nodes:
            # This node is part of a path that has already been traversed.
            continue

        paths = [[parent]]
        while paths:
            path = paths.pop()

            parent = path[-1]

            try:
                children = parent_children[parent]
            except KeyError:
                continue

            for child in children:
                # Keeping a set of the path nodes, for O(1) lookups at the
                # expense of more memory and complexity, actually makes speed
                # worse. (Due to construction of sets.)
                # This is O(N).
                if child in path:
                    # This is a cycle.
                    cycle = path[path.index(child):]
                    # Check that this is not a dup cycle.
                    is_dup = False
                    for other_cycle in cycles:
                        if is_rotated(other_cycle, cycle):
                            is_dup = True
                            break
                    if not is_dup:
                        cycles.append(cycle)
                        yield cycle
                else:
                    # Push this new path onto the 'stack'.
                    # This is probably the most expensive part of the algorithm
                    # (a list copy).
                    paths.append(path + [child])
                    # Mark the node as visited.
                    visited_nodes.add(child)


if __name__ == '__main__':
    # Run the doctest tests.
    import sys
    import doctest
    doctest.testmod(sys.modules['__main__'])



"""
Support for the msgpack-based binary format
"""
from msgpack import Unpacker, Packer
from collections import defaultdict, OrderedDict

class MsgpackReader(object):
    def __init__(self, f):
        self.f = f
        self.unpacker = Unpacker(f)

def schema_to_dict(schema):
    d = OrderedDict()
    for attr in schema.attributes:
        kind = attr.get_kind()
        if kind == 'ENUM':
            d[attr.name] = ['ENUM', list(attr.alphabet)]
        else:
            d[attr.name] = kind
    return d

def objects_to_packed(doc, objs, schema, want_span):
    names = []
    data = []
    names.append(':id')
    data.append([getattr(obj, 'xml_id', None) for obj in objs])
    if want_span:
        n = len(objs)
        span_array = [0] * (2*n)
        last_offset = 0
        for i, obj in enumerate(objs):
            span_array[i] = obj.span[0] - last_offset
            span_array[n+1] = obj.span[-1] - obj.span[0]
            last_offset = obj.span[-1]
        names.append(':span')
        data.append(span_array)
    for attr in schema.attributes:
        att_seq = []
        any_filled = False
        for obj in objs:
            val = getattr(obj, attr.prop_name, None)
            if val is not None:
                any_filled = True
                att_seq.append(attr.map_attr(val, doc))
            else:
                att_seq.append(None)
        if any_filled:
            names.append(attr.name)
            data.append(att_seq)
    return [len(objs), names, data]


class MsgpackWriter(object):
    def __init__(self, f):
        self.f = f
        self.packer = Packer()

    def pack(self, obj):
        self.f.write(self.packer.pack(obj))

    def write_document(self, doc):
        """
        Writes a complete EXML document

        :param doc Document: the document to write
        """
        packer = self.packer
        self.f.write(
            packer.pack_array_header(3) +
            packer.pack('exml1'))
        self.write_schema(doc)
        self.write_chunks(doc)

    def write_schema(self, doc):
        """
        Writes the schema part of an EXML document

        :param doc: the document to write
        :type doc Document:
        """
        d = {}
        d['word'] = schema_to_dict(doc.t_schema)
        for schema in doc.schemas:
            d[schema.name] = schema_to_dict(schema)
        self.pack(d)

    def write_chunks(self, doc):
        self.f.write(self.packer.pack_array_header(1))
        self.write_chunk(doc, 0, None)

    def write_chunk(self, doc, start, end):
        if end is None:
            end = len(doc.words)
        self.f.write(self.packer.pack_array_header(2))
        # write terminals
        self.pack(objects_to_packed(doc, doc.w_objs, doc.t_schema, False))
        self.write_markables(doc)

    def write_markables(self, doc):
        markables_by_layer = defaultdict(list)
        for posn in sorted(doc.markables_by_start):
            for (mlevel, obj) in doc.markables_by_start[posn]:
                markables_by_layer[mlevel.name].append(obj)
        ne_levels = []
        for schema in doc.schemas:
            if schema.name not in markables_by_layer:
                continue
            ne_levels.append(schema.name)
        self.f.write(self.packer.pack_map_header(len(ne_levels)))
        for schema in doc.schemas:
            if schema.name not in markables_by_layer:
                continue
            objs = markables_by_layer[schema.name]
            self.pack(schema.name)
            self.pack(objects_to_packed(doc, objs, schema, True))

if __name__ == '__main__':
    import sys
    from exmldoc import load
    doc = load(sys.argv[1])
    with open(sys.argv[2], 'wb') as f_out:
        wr = MsgpackWriter(f_out)
        wr.write_document(doc)


from __future__ import print_function
import sys
import re

unwanted_mrg = re.compile(r"([^A-Za-z0-9\x80-\xff\-_])")


def escape_mrg(string):
    if not string:
        return ''
    return unwanted_mrg.sub(r"\\\1", string)


def bottomup_enumeration(nodes):
    # print "bottomup_enumeration: %s"%(nodes)
    for n in nodes:
        for n1 in bottomup_enumeration(n.children):
            yield n1
        yield n


def descendants(node):
    for n in node.children:
        yield n
        for n1 in descendants(n):
            yield n1
    return


def determine_tokenspan(node):
    if not node.isTerminal():
        if not node.children:
            raise ValueError('NT Node %s#%s has no children' % (
                node.cat, node.id))
        node.start = min(map(lambda x: x.start, node.children))
        node.end = max(map(lambda x: x.end, node.children))


class Tree(object):
    '''
    a Tree object represents a syntax tree.
    You can access roots (nodes with no parent node) and
    the leaves (preterminal nodes) through its attributes.

    .. py:attribute:: roots

       The roots of the tree.

    .. py:attribute:: terminals

       The leaves/pre-terminals of the tree

    .. py:attribute:: encoding

       The encoding for any (non-unicode) string values in the tree's nodes
    '''
    __slots__ = ['node_table', 'roots', 'terminals', 'encoding',
                 '__dict__']

    def __getstate__(self):
        return (self.node_table,
                self.roots,
                self.terminals,
                self.__dict__)

    def __setstate__(self, state):
        self.node_table, self.roots, self.terminals, self.__dict__ = state

    def __init__(self):
        self.node_table = {}
        self.roots = []
        self.terminals = []

    def __iter__(self):
        return iter(self.roots)

    def bottomup_enumeration(self):
        '''
        returns a sequence of all (nonterminal as well
        as terminal) nodes in the tree, in left-to-right,
        bottom-up order
        '''
        return bottomup_enumeration(self.roots)

    def topdown_enumeration(self):
        '''
        returns a sequence of all (nonterminal and terminal)
        nodes in the tree, in left-to-right, top-down order
        '''
        for n in self.roots:
            yield n
            for n1 in descendants(n):
                yield n1
        return

    def determine_tokenspan_all(self):
        "determines the tokenspan for all nodes and sorts children accordingly"
        def by_pos(a, b):
            return cmp(a.start, b.start)
        try:
            for node in self.bottomup_enumeration():
                determine_tokenspan(node)
                node.children.sort(by_pos)
        except:
            print("In tree: %s" %
                  (getattr(self, 'sent_no', '???'),), file=sys.stderr)
            raise
        self.roots.sort(by_pos)

    def check_roots(self):
        for n in self.roots:
            assert n.parent is None, (repr(n), getattr(self, 'sent_no', None))
            self.check_nodes(n, [])

    def check_nodes(self, node, parents):
        if node.parent is None:
            assert parents == [], (repr(node), getattr(self, 'sent_no', None))
        else:
            assert node.parent == parents[-1], (repr(node),
                                                getattr(self, 'sent_no', None))

        parents.append(node)
        for n in node.children:
            assert not n in parents
            self.check_nodes(n, parents)
        del parents[-1]

    def check_node_table(self):
        for k, n in self.node_table.iteritems():
            assert n.id == k

    def renumber_ids(self, nodes=None, start=500):
        """gives ids to all nonterminal nodes."""
        pos = start
        if nodes == None:
            # renumber all IDs and clear out node_table
            self.node_table = {}
            nodes = self.roots
        for n in nodes:
            if not n.isTerminal():
                # print "Renumber %r: entering %s, pos=%d"%(n.id,pos)
                pos = 1 + self.renumber_ids(n.children, pos)
                #sys.stderr.write("Renumber %r: %s => %d\n"%(n,n.id,pos))
                n.id = "%s" % pos
                self.node_table[n.id] = n
        return pos

    def check_nodetable(self):
        for key in self.node_table:
            if self.node_table[key].id != key:
                raise "Nodetable: node %s(%r) has id %s" % (key,
                                                            self.node_table[key], self.node_table[key].id)
            assert self.node_table[key].id == key
            if self.node_table[key].parent == None:
                assert self.node_table[key] in self.roots
            else:
                parent = self.node_table[key].parent
                assert self.node_table[parent.id] == parent

    def discontinuity(self, nodes, index, sent_node):
        """returns True iff there is a discontinuity between
        the Nth and the N+1th member of nodes, ignoring
        punctuation and parentheses."""
        if nodes[index].end == nodes[index + 1].start:
            return False
        sys.stderr.write('Looking for a discontinuity between %r and %r' % (
            self.terminals[nodes[index].end],
            self.terminals[nodes[index + 1].start]))
        for n in self.terminals[nodes[index].end:nodes[index + 1].start]:
            n1 = n
            while n1 != None:
                if n1 == sent_node:
                    return True
                n1 = n1.parent
        return False


# abstract base class for all nodes
class Node(object):

    def __init__(self, cat):
        self.id = None
        self.start = -1
        self.end = -1
        self.cat = cat
        self.children = []
        self.parent = None

    def add_at(self, node, pos):
        self.children[pos:pos] = [node]
        node.set_parent(self)

    def append(self, node):
        self.children.append(node)
        node.set_parent(self)

    def insert(self, node):
        "inserts a node at the appropriate position"
        node.set_parent(self)
        for (i, n) in enumerate(self.children):
            if (n.start >= node.start):
                self.children[i:i] = [node]
                return
        self.append(node)

    def set_parent(self, parent):
        self.parent = parent


class NontermNode(Node):
    '''
    Node class for nonterminal node

    The exact set of attributes for a nonterminal
    is use-dependent, but the core set of these consists in

    .. py:attribute:: cat

       the node category (e.g. NP)

    .. py:attribute:: edge_label

       the grammatical function/edge label of this node
       (e.g. subject, modifier)

    .. py:attribute:: children

       a list of the children of this node, sorted by starting
       position

    For many uses, the following attributes are also relevant:

    .. py:attribute:: start

       the start position (usually inferred through determine_tokenspans_all
       as the leftmost position of any terminal in its yield).

    .. py:attribute:: end

       the end position, as one more than the rightmost position of
       any terminal in its yield

    .. py:attribute:: xml_id

       the identifier that this nodes gets in any XML-based formats
       such as EXML, TigerXML or PML
    '''

    def __init__(self, cat, edge_label=None):
        Node.__init__(self, cat)
        self.edge_label = edge_label
        self.attr = '--'

    def __repr__(self):
        stuff = ''
        if hasattr(self, 'xml_id'):
            stuff += '#' + self.xml_id
        stuff += ' at ' + hex(id(self))
        return '<%s.%s%s>' % (self.cat, self.edge_label, stuff)

    def isTerminal(self):
        return False

    def __str__(self):
        return '<NonTerm %s #%s>' % (self.cat, self.id)

    def to_penn(self):
        s = self.cat
        s = s.replace('(', '-LRB-')
        s = s.replace(')', '-RRB-')
        a = "(%s " % (s,)
        a += ' '.join(map(lambda x: x.to_penn(), self.children))
        a += ")"
        return a

    def to_full(self, wanted_attrs):
        pairs = []
        for key in wanted_attrs:
            pairs.append('%s=%s' %
                         (key, escape_mrg(str(getattr(self, key, '--')))))
        a = "(%s" % (escape_mrg(self.cat))
        if pairs:
            a = a + "=#i[%s]" % (' '.join(pairs))
        a += " %s)" % (' '.join(map(lambda x: x.to_full(wanted_attrs), self.children)),
                       )
        return a


class TerminalNode(Node):

    '''
    Node class for a preterminal node.

    These usually have the following attributes:

    .. py:attribute:: word

       the surface form of the word of this terminal

    .. py:attribute:: cat

       the POS tag (syntactic category)

    .. py:attribute:: morph

       the morphological tag

    .. py:attribute:: lemma

       the lemma of this node

    .. py:attribute:: start

       the position of this node

    .. py:attribute:: xml_id

       the XML id of this node

    If this node is part of a dependency tree, it will usually also have
    the attributes:

    .. py:attribute:: syn_parent

       the governor of this word (i.e., a reference to another TerminalNode)

    .. py:attribute:: syn_label

       the dependency label of this word

    '''

    def __init__(self, cat, word, edge_label=None, morph=None):
        Node.__init__(self, cat)
        self.word = word
        self.edge_label = edge_label
        self.morph = morph

    def __repr__(self):
        if hasattr(self, 'xml_id'):
            stuff = '#' + self.xml_id
        else:
            stuff = '(%d)' % (self.start)
        return '<%s/%s%s at %s>' % (self.word, self.cat, stuff, hex(id(self)))

    def isTerminal(self):
        return True

    def to_penn(self):
        cat = self.cat
        cat = cat.replace('(', '-LRB-')
        cat = cat.replace(')', '-RRB-')
        word = self.word
        word = word.replace('(', '-LRB-')
        word = word.replace(')', '-RRB-')
        return "(%s %s)" % (self.cat, self.word)

    def to_full(self, wanted_attrs):
        pairs = []
        for key in wanted_attrs:
            pairs.append('%s=%s' %
                         (key, escape_mrg(str(getattr(self, key, '--')))))
        a = "(%s" % (escape_mrg(self.cat),)
        if pairs:
            a = a + "=#i[%s]" % (' '.join(pairs))
        a += " %s)" % (escape_mrg(self.word),)
        return a



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
exmldoc contains code related to the ExportXMLv2 format, including
the object model for corpus schemata, functions to write out corpora
in XML and JSON formats, and functions to read single documents or corpora
in JSON format.
"""
from __future__ import print_function

import sys
import re
from isounidecode.unidecode import unidecode
from collections import OrderedDict, defaultdict
import gzip
from sortedcontainers import SortedDict
from xml.sax.saxutils import quoteattr, escape
import simplejson as json

try:
    import lxml.etree as etree
    have_lxml = True
except ImportError:
    import xml.etree.cElementTree as etree
    have_lxml = False



from . import tree
from .topsort import topsort
from .alphabet import PythonAlphabet

__version__ = "2014-07-08"
__author__ = "Yannick Versley / Univ. Heidelberg"

if sys.version_info.major >= 3:
    intern = sys.intern
    xrange = range
    unicode = str
    islice = slice
else:
    from itertools import izip, islice

class _EmptyClass:
    pass


QNAME_XML_ID = '{http://www.w3.org/XML/1998/namespace}id'

TEMP_ID = 0


def create_id(prefix, alphabet):
    n0 = len(alphabet)
    n = n0
    while alphabet['%s%d' % (prefix, n)] < n0:
        n += 1
    return '%s%d' % (prefix, n)


class TextAttribute:

    """
    an attribute with a string value
    """

    def __init__(self, name, prop_name=None, default_val=None):
        self.name = name
        if prop_name is None:
            self.prop_name = name
        else:
            self.prop_name = prop_name
        self.default_val = default_val

    def map_attr(self, val, doc):
        if val == self.default_val:
            return None
        return val

    def unmap_attr(self, val, doc, encoding):
        return unmunge_xml(val, encoding)

    def get_updown(self, obj, doc, result):
        pass

    def get_kind(self):
        return 'STRING'

    def describe_schema(self, f, encoding=None):
        open_tag(f, 'text-attr', [('name', self.name)],
                 indent=2, encoding=encoding)
        f.write('/>\n')


class EnumAttribute(TextAttribute):
    """
    an attribute with a value that comes from an
    enumeration. In the XML format, all declared enumeration
    values are listed in the schema header.
    """

    def __init__(self, name, **kw):
        TextAttribute.__init__(self, name, **kw)
        self.alphabet = PythonAlphabet()
        self.descriptions = {}

    def add_item(self, name, description=None):
        self.alphabet[name]
        if description is not None:
            self.descriptions[name] = description

    def describe_schema(self, f, encoding=None):
        open_tag(f, 'enum-attr', [('name', self.name)],
                 indent=2, encoding=encoding)
        f.write('>\n')
        for val in self.alphabet.words:
            atts = [('name', val)]
            if val in self.descriptions:
                atts.append(('description', self.descriptions[val]))
            open_tag(f, 'val', atts,
                     indent=3, encoding=encoding)
            f.write('/>\n')
        f.write('  </enum-attr>\n')

    def get_kind(self):
        return 'ENUM'


class RefAttribute:
    """
    an attribute containing a reference to another
    object (terminal or markable).
    """

    def __init__(self, name, prop_name=None, restriction=None,
                 restrict_target=None):
        self.name = name
        if prop_name is None:
            self.prop_name = self.name
        else:
            self.prop_name = prop_name
        if restriction is None or restriction == 'none':
            self.restriction = None
        else:
            self.restriction = intern(restriction)
        self.restrict_target = restrict_target

    def map_attr(self, val, doc):
        return doc.get_obj_id(val)

    def unmap_attr(self, val, doc, encoding=None):
        return doc.object_by_id[val]

    def get_updown(self, obj, doc, result):
        if self.restriction is 'down':
            other_obj = getattr(obj, self.prop_name)
            if other_obj is not None:
                result.append((doc.get_obj_id(obj),
                               doc.get_obj_id(other_obj)))
        elif self.restriction is 'up':
            other_obj = getattr(obj, self.prop_name)
            if other_obj is not None:
                result.append((doc.get_obj_id(other_obj),
                               doc.get_obj_id(obj)))

    def describe_schema(self, f, encoding=None):
        open_tag(f, 'node-ref', [('name', self.name)],
                 indent=2, encoding=encoding)
        f.write('/>\n')

    def get_kind(self):
        return 'REF'

class IDRefAttribute:

    """
    an attribute containing the string ID of another
    terminal or node
    """

    def __init__(self, name, prop_name=None, restriction=None,
                 restrict_target=None):
        self.name = name
        if prop_name is None:
            self.prop_name = self.name
        else:
            self.prop_name = prop_name
        if restriction is None or restriction == 'none':
            self.restriction = None
        else:
            self.restriction = sys.intern(restriction)
        self.restrict_target = restrict_target

    def map_attr(self, val, doc):
        return val

    def unmap_attr(self, val, doc, encoding):
        return val

    def get_updown(self, obj, doc, result):
        if self.restriction is 'down':
            other = getattr(obj, self.prop_name)
            if other is not None:
                result.append((doc.get_obj_id(obj),
                               other))
        elif self.restriction is 'up':
            other = getattr(obj, self.prop_name)
            if other is not None:
                result.append((other,
                               doc.get_obj_id(obj)))

    def describe_schema(self, f, encoding=None):
        open_tag(f, 'node-ref', [('name', self.name)],
                 indent=2, encoding=encoding)
        f.write('/>\n')

    def get_kind(self):
        return 'REF'

class MarkableSchema:

    """
    a document schema that describes the mapping from Python
    objects and attributes to nodes in the serialization.
    """

    def __init__(self, name, cls=None, kind="Markable"):
        self.name = name
        self.attributes = []
        self.init_attrs = []
        self.edges = []
        self.interfaces = set()
        self.cls = cls
        self.locality = None
        self.suffix = kind

    def __repr__(self):
        return '<MarkableSchema %s %s>'%(self.name, id(self))

    def serialize_object(self, obj, doc, force_ids=True):
        span = obj.span
        attr_d = OrderedDict()
        if force_ids or hasattr(obj, 'xml_id'):
            oid = doc.get_obj_id(obj)
            attr_d['xml:id'] = oid
        for att in self.attributes:
            if hasattr(obj, att.prop_name):
                v = getattr(obj, att.prop_name)
                if v is not None:
                    v_txt = att.map_attr(v, doc)
                    if v_txt is not None:
                        attr_d[att.name] = v_txt
        edges = []
        for edge_schema in self.edges:
            edgelist = edge_schema.get_edges(obj, doc)
            for edgevals in edgelist:
                attr_e = OrderedDict()
                for (att, val) in izip(edge_schema.attributes, edgevals):
                    if val is not None:
                        attr_e[att.name] = att.map_attr(val, doc)
                edges.append((edge_schema.name, attr_e))
        return (span, self.name, attr_d, edges)

    def make_json(self, obj, doc):
        oid = doc.get_obj_id(obj)
        attrs = {'_id': oid, 'span': obj.span}
        for att in self.attributes:
            if hasattr(obj, att.prop_name):
                v = getattr(obj, att.prop_name)
                if v is not None:
                    attrs[att.name] = munge_json(v)
        for edge in self.edges:
            v = edge.get_edges(obj, doc)
            if v:
                attrs[edge.name] = munge_json(v)
        return attrs

    def create_from_json(self, obj, doc):
        args = []
        for att in self.init_attrs:
            v = None
            try:
                v = unmunge_json(obj[att.prop_name], doc)
            except KeyError:
                pass
            args.append(v)
        try:
            m = self.cls(*args)
        except TypeError:
            print("Cannot instantiate %s" % (self.cls), file=sys.stderr)
            raise
        m.xml_id = obj['_id']
        m.span = obj['span']
        doc.object_by_id[m.xml_id] = m
        return m

    def create_from_xml(self, elem, doc, encoding=None):
        args = []
        for att in self.init_attrs:
            v = None
            try:
                v = unmunge_xml(elem.attrib[att.prop_name], encoding)
            except KeyError:
                pass
            args.append(v)
        try:
            m = self.cls(*args)
        except TypeError:
            print("Cannot instantiate %s" % (self.cls), file=sys.stderr)
            raise
        try:
            m.xml_id = elem.attrib[QNAME_XML_ID]
        except KeyError:
            obj_id = doc.assign_temp_id(m)
            elem.attrib[QNAME_XML_ID] = obj_id
        doc.object_by_id[m.xml_id] = m
        return m

    def fill_from_json(self, obj, attrs, doc):
        for att in self.attributes:
            if att.name in attrs:
                setattr(obj, att.prop_name, unmunge_json(attrs[att.name], doc))
        for edge in self.edges:
            if edge.name in attrs:
                edge.set_edges(obj, unmunge_json(attrs[edge.name], doc), doc)

    def fill_from_xml(self, obj, elem, doc, encoding):
        for att in self.attributes:
            if att.name in elem.attrib:
                setattr(obj, att.prop_name,
                        att.unmap_attr(elem.attrib[att.name], doc, encoding))
        if 'span' in elem.attrib:
            obj.span = decode_span(elem.attrib['span'], doc)

    def get_updown(self, obj, doc, result):
        for att in self.attributes:
            att.get_updown(obj, doc, result)
        for edge in self.edges:
            edge.get_updown(obj, doc, result)

    def describe_schema(self, f, edges, encoding=None):
        attrs = [('name', self.name)]
        if self.locality is not None:
            attrs.append(('locality', self.locality))
        open_tag(f, 'node', attrs, 1, encoding=encoding)
        f.write('>\n')
        for att in self.attributes:
            att.describe_schema(f, encoding=encoding)
        f.write(' </node>\n')
        for edge_schema in self.edges:
            if edge_schema.name not in edges:
                edges[edge_schema.name] = [edge_schema, [self.name]]
            else:
                edges[edge_schema.name][1].append(self.name)

    def check_interface(self, name, descr):
        result = []
        for (tp, part) in descr:
            try:
                if tp == 'attr':
                    att = self.attribute_by_name(part)
                elif tp == 'edge':
                    att = self.edge_by_name(part)
                else:
                    assert False, tp
                result.append((tp, att))
            except KeyError:
                print("No %s" % (part,), file=sys.stderr)
                return False
        self.interfaces.add(name)
        return result

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def edge_by_name(self, name):
        for edge in self.edges:
            if edge.name == name:
                return edge
        raise KeyError(name)

    def add_edge(self, edge):
        self.edges.append(edge)


class SecondaryEdges:

    def __init__(self, name):
        self.name = name
        self.alphabet = PythonAlphabet()
        self.attributes = [EnumAttribute('cat'),
                           RefAttribute('parent',
                                        restrict_target=['word', 'node'])]
        self.suffix = 'Edge'
        self.descriptions = {}

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def get_edges(self, obj, doc):
        edges = []
        if hasattr(obj, 'secedge') and obj.secedge is not None:
            for secedge in obj.secedge:
                edges.append([secedge[0], secedge[1]])
        return edges

    def set_edges(self, obj, vals, doc):
        obj.secedge = vals

    def get_updown(self, obj, doc, result):
        pass

    def add_item(self, name, description=None):
        self.alphabet[name]
        if description is not None:
            self.descriptions[name] = description


def decode_span(s_span, doc):
    # TODO reject invalid word_ids
    result = []
    for s_part in s_span.split(','):
        if '..' in s_part:
            (s_start, s_end) = s_part.split('..', 2)
            start = doc.word_ids[s_start]
            end = doc.word_ids[s_end] + 1
            result += [start, end]
        else:
            start = doc.word_ids[s_part]
            result += [start, start + 1]
    return result


def open_tag(f, name, items, indent=0, encoding=None):
    f.write(' ' * indent)
    f.write('<%s' % (name,))
    for k, v in items:
        if v is None:
            continue
        if isinstance(v, unicode):
            if encoding is not None:
                f.write(' %s="%s"' %
                        (k, escape(v.encode(encoding), {'"': "&quot;"})))
            else:
                f.write(' %s=%s'%(
                    k, quoteattr(v).encode('ascii', 'xmlcharrefreplace')))
        else:
            f.write(' %s=%s' % (k, quoteattr(v)))


class ChildEdges:

    def __init__(self, name):
        self.name = name
        self.attributes = [RefAttribute('target'),
                           TextAttribute('label')]
        self.suffix = 'Edge'

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def put_edges(self, obj, doc, edges):
        for n in obj.children:
            attr_d = OrderedDict()
            attr_d['target'] = doc.get_obj_id(n)
            attr_d['label'] = n.edge_label
            edges.append((self.name, attr_d))

    def get_updown(self, obj, doc, result):
        obj_id = doc.get_obj_id(obj)
        for n in obj.children:
            result.append((doc.get_obj_id(n),
                           obj_id))


class ReferenceEdges(object):

    def __init__(self, name):
        self.name = name
        self.attributes = [EnumAttribute('type'),
                           IDRefAttribute('target',
                                          restrict_target=['word', 'node'])]
        self.suffix = 'Edge'

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def get_edges(self, obj, doc):
        info = getattr(obj, 'anaphora_info', None)
        if info is not None:
            tgt = None
            if info[0] == 'split_antecedent':
                return []
            elif info[0] not in ['expletive', 'inherent_reflexive']:
                tgt = ' '.join(info[1])
                # tgt=info[1]
            return [[info[0], tgt]]
        else:
            return []

    def set_edges(self, obj, vals, doc):
        assert len(vals) == 1
        kind = vals[0][0]
        if len(vals[0]) > 1 and vals[0][1] is not None:
            targets = vals[0][1].split(' ')
        else:
            targets = None
        obj.anaphora_info = [kind, targets]

    def get_updown(self, obj, doc, result):
        pass


class GenericEdges(object):
    def __init__(self, name, prop_name=None):
        self.name = name
        if prop_name is None:
            self.prop_name = name
        else:
            self.prop_name = prop_name
        self.attributes = []
        self.suffix = 'Edge'

    def find_attribute(self, name):
        for i, att in enumerate(self.attributes):
            if att.name == name:
                return i, att
        raise KeyError(name)

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def get_edges(self, obj, doc):
        if hasattr(obj, self.prop_name):
            return getattr(obj, self.prop_name)
        else:
            return []

    def set_edges(self, obj, vals, doc):
        setattr(obj, self.prop_name, vals)

    def get_updown(self, obj, doc, result):
        pass


class SplitRefEdges(object):

    def __init__(self, name):
        self.name = name
        self.attributes = [EnumAttribute('type'),
                           TextAttribute('target')]
        self.suffix = 'Edge'

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def get_edges(self, obj, doc):
        info = getattr(obj, 'anaphora_info', None)
        if info is not None:
            tgt = None
            if info[0] == 'split_antecedent':
                tgt = ' '.join(info[1])
                return [[info[0], tgt]]
            else:
                return []
        else:
            return []

    def set_edges(self, obj, vals, doc):
        assert len(vals) == 1
        kind = vals[0][0]
        targets = vals[0][1].split(' ')
        obj.anaphora_info = [kind, targets]

    def get_updown(self, obj, doc, result):
        pass


def munge_json(obj):
    if isinstance(obj, list) or isinstance(obj, tuple):
        return map(munge_json, obj)
    elif isinstance(obj, int) or isinstance(obj, unicode):
        return obj
    elif isinstance(obj, str):
        return obj.decode('ISO-8859-15')
    elif hasattr(obj, 'xml_id'):
        return {'_id': obj.xml_id}
    elif obj is None:
        return None
    assert False, obj


# TODO: \u0219 \u2022
uc_dash = re.compile(u'[\u2010\u2012\u2013\u2014\u2015\u2212]', re.U)
uc_squo = re.compile(u'[\u2018\u2019\u201a\u2032\u02b9\u2039\u203a]', re.U)
uc_dquo = re.compile(u'[\u201c\u201d\u201e\u2033\u02ba]', re.U)
uc_bullet = re.compile(u'[\u2022\u2020\u2021]', re.U)


def normalize_string(s):
    s = uc_dash.sub(u'-', s)
    s = uc_squo.sub(u"'", s)
    s = uc_dquo.sub(u'"', s)
    s = uc_bullet.sub(u'*', s)
    return s


def unmunge_json(obj, doc):
    if isinstance(obj, list) or isinstance(obj, tuple):
        return [unmunge_json(x, doc) for x in obj]
    elif isinstance(obj, int):
        return obj
    elif isinstance(obj, unicode):
        return unidecode(normalize_string(obj), 'iso8859-1')
    elif isinstance(obj, str):
        return obj
    elif obj is None:
        return obj
    elif '_id' in obj:
        return doc.object_by_id[obj['_id']]
    assert False, obj


def unmunge_xml(obj, encoding):
    if isinstance(obj, unicode):
        if encoding is None:
            return obj
        elif encoding == 'UTF-8':
            return obj.encode(encoding)
        else:
            return unidecode(normalize_string(obj), encoding)
    elif isinstance(obj, bytes):
        if encoding is None:
            return unicode(obj)
        return obj


def to_string(s):
    if isinstance(s, unicode):
        return unidecode(normalize_string(s), 'iso8859-1')
    elif isinstance(s, bytes):
        return s


class TerminalSchema(object):

    """
    The class for Terminal schema objects
    """

    def __init__(self, name, cls):
        self.name = name
        self.attributes = []
        self.edges = []
        self.interfaces = set()
        self.cls = cls

    def serialize_terminal(self, obj, doc, force_ids=True):
        # span=obj.span
        attr_d = OrderedDict()
        if force_ids or hasattr(obj, 'xml_id'):
            oid = doc.get_obj_id(obj)
            attr_d['xml:id'] = oid
        for att in self.attributes:
            if hasattr(obj, att.prop_name):
                v = getattr(obj, att.prop_name)
                if v is not None:
                    v_txt = att.map_attr(v, doc)
                    if v_txt is not None:
                        attr_d[att.name] = v_txt
        edges = []
        for edge_schema in self.edges:
            edgelist = edge_schema.get_edges(obj, doc)
            for edgevals in edgelist:
                attr_e = OrderedDict()
                for (att, val) in izip(edge_schema.attributes, edgevals):
                    if val is not None:
                        attr_e[att.name] = att.map_attr(val, doc)
                edges.append((edge_schema.name, attr_e))
        return (self.name, attr_d, edges)

    def make_json(self, obj, doc):
        oid = doc.get_obj_id(obj)
        attrs = {'_id': oid}
        for att in self.attributes:
            if hasattr(obj, att.prop_name):
                v = getattr(obj, att.prop_name)
                if v is not None:
                    attrs[att.name] = munge_json(v)
        for edge in self.edges:
            v = edge.get_edges(obj, doc)
            if v:
                attrs[edge.name] = munge_json(v)
        return attrs

    def create_from_json(self, attrs, doc):
        # special case: cat/pos is filled or None, word/form MUST be filled
        obj = self.cls(attrs.get('pos'), to_string(attrs['form']))
        if '_id' in attrs:
            s_id = attrs['_id']
            obj.xml_id = s_id
            doc.object_by_id[s_id] = obj
        return obj

    def create_from_xml(self, node, doc, encoding=None):
        obj = self.cls(node.attrib.get('pos', None),
                       unmunge_xml(node.attrib['form'], encoding))
        if QNAME_XML_ID in node.attrib:
            obj.xml_id = node.attrib[QNAME_XML_ID]
        else:
            obj_id = doc.assign_temp_id(obj)
            node.attrib[QNAME_XML_ID] = obj_id
        return obj

    def fill_from_json(self, obj, attrs, doc):
        for att in self.attributes:
            if att.name in attrs:
                setattr(obj, att.prop_name, unmunge_json(attrs[att.name], doc))
        for edge in self.edges:
            if edge.name in attrs:
                edge.set_edges(obj, unmunge_json(attrs[edge.name], doc), doc)

    def fill_from_xml(self, obj, node, doc, encoding):
        for att in self.attributes:
            if att.name in node.attrib:
                setattr(obj, att.prop_name,
                        att.unmap_attr(node.attrib[att.name], doc, encoding))

    def describe_schema(self, f, edges, encoding=None):
        open_tag(f, 'tnode', [('name', self.name)], 1,
                 encoding=encoding)
        f.write('>\n')
        for att in self.attributes:
            att.describe_schema(f, encoding=encoding)
        f.write(' </tnode>\n')
        for edge_schema in self.edges:
            if edge_schema.name not in edges:
                edges[edge_schema.name] = [edge_schema, [self.name]]
            else:
                edges[edge_schema.name][1].append(self.name)

    def check_interface(self, name, descr):
        result = []
        for (tp, part) in descr:
            try:
                if tp == 'attr':
                    att = self.attribute_by_name(part)
                elif tp == 'edge':
                    att = self.edge_by_name(part)
                else:
                    assert False, tp
                result.append((tp, att))
            except KeyError:
                print("No %s" % (part,), file=sys.stderr)
                return False
        self.interfaces.add(name)
        return result

    def attribute_by_name(self, att_name):
        for att in self.attributes:
            if att.name == att_name:
                return att
        raise KeyError(att_name)

    def add_attribute(self, att):
        self.attributes.append(att)

    def edge_by_name(self, name):
        for edge in self.edges:
            if edge.name == name:
                return edge
        raise KeyError(name)

    def add_edge(self, edge):
        self.edges.append(edge)

class SortedBag(SortedDict):
    def __missing__(self, key):
        val = list()
        self[key] = val
        return val

class Document:

    """
    represents an ExportXMLv2 document, including
    words, terminals and markables on different
    annotation layers.
    """
    # TODO use a SortedDictionary to maintain markables efficiently
    # either one per level or one globally
    # http://www.grantjenks.com/docs/sortedcontainers/sorteddict.html

    def __init__(self, t_schema, schemas):
        """
        Creates a Document with the specified annotation layers
        :param TerminalSchema t_schema: the schema for terminal nodes
        :param List[MarkableSchema] schemas: the schemas for each annotation layer 
        """
        self.t_schema = t_schema
        self.schemas = schemas
        self.schema_by_class = {}
        self.object_by_id = {}
        self.temp_ids = {}
        # self.basedata=BaseData()
        self.words = []
        self.w_objs = []
        self.word_attr = 'word'
        self.markables_by_start = SortedBag()
        self.node_objs = defaultdict(list)
        self.word_ids = PythonAlphabet()
        self.interface_classes = defaultdict(list)
        self.interface_attrs = {}
        if t_schema.cls is not None:
            self.schema_by_class[t_schema.cls] = t_schema
        for schema in schemas:
            if schema.cls is not None:
                self.schema_by_class[schema.cls] = schema

    def __len__(self):
        return len(self.words)

    def size(self):
        return len(self.words)

    def add_schemas(self, schemas):
        self.schemas += schemas
        for schema in schemas:
            if schema.cls is not None:
                self.schema_by_class[schema.cls] = schema

    def schema_by_name(self, name):
        for schema in self.schemas:
            if schema.name == name:
                return schema
        raise KeyError

    def add_interface(self, name, spec):
        result = self.t_schema.check_interface(name, spec)
        if result:
            self.interface_classes[name].append('word')
            self.interface_attrs[name] = result
        for schema in self.schemas:
            result = schema.check_interface(name, spec)
            if result:
                self.interface_classes[name].append(schema.name)
                self.interface_attrs[name] = result

    def get_interfaces(self, name):
        """returns all the interfaces that the level called name fulfills"""
        result = []
        for k, v in self.interface_classes.items():
            print(name, k, v, (name in v), file=sys.stderr)
            if name in v:
                result.append(k)
        print(result, file=sys.stderr)
        return result

    def common_interface(self, names):
        if names[0] == 'word':
            cands = self.t_schema.interfaces
        else:
            cands = self.schema_by_name(names[0]).interfaces
        for name in names[1:]:
            cands.intersection_update(self.schema_by_name(name).interfaces)
        if cands:
            return sorted(cands)[0]
        else:
            return None

    def get_obj_id(self, obj):
        if hasattr(obj, 'xml_id'):
            return obj.xml_id
        else:
            mlevel = self.mlevel_for_class(type(obj))
            if mlevel is not None:
                n = mlevel.name
            else:
                n = 'x'
            k = '%s_%s' % (n, id(obj))
            obj.xml_id = k
            self.object_by_id[k] = obj
            return k

    def assign_temp_id(self, obj):
        global TEMP_ID
        TEMP_ID += 1
        k = '__tmp_%d'%(TEMP_ID,)
        obj.xml_id = k
        self.object_by_id[k] = obj
        self.temp_ids[k] = obj
        return k

    def clear_temp_id(self, obj):
        if hasattr(obj, 'xml_id') and obj.xml_id in self.temp_ids:
            k = obj.xml_id
            del obj.xml_id
            del self.object_by_id[k]
            del self.temp_ids[k]

    def clear_temp_ids(self):
        for k in self.temp_ids:
            obj = self.temp_ids[k]
            del obj.xml_id
            del self.object_by_id[k]
        self.temp_ids = {}

    def add_terminal(self, w_obj):
        val = self.word_ids[self.get_obj_id(w_obj)]
        assert val == len(self.words), (val, w_obj.xml_id,
                                        len(self.words), self.words[val - 2:val + 2], self.words[-2:])
        self.words.append(getattr(w_obj, self.word_attr))
        self.w_objs.append(w_obj)

    def replace_terminal(self, posn, w_obj):
        w_obj.xml_id = self.word_ids.get_sym(posn)
        assert self.words[posn] == getattr(
            w_obj, self.word_attr), (self.words[posn], getattr(w_obj, self.word_attr))
        self.w_objs[posn] = w_obj
        self.object_by_id[w_obj.xml_id] = w_obj

    def mlevel_for_class(self, cls):
        try:
            return self.schema_by_class[cls]
        except KeyError:
            for k in cls.__bases__:
                result = self.mlevel_for_class(k)
                if result is not None:
                    return result
        return None

    def reorder_updown(self, objs):
        # 1. extract up/down graph
        # TODO: add precedence for "locality"-type things
        edges = []
        objs_dict = {}
        result = []
        objs_by_level = {}
        edges_len = 0
        temp_names = []
        for (ml, obj) in objs:
            if not hasattr(obj, 'xml_id'):
                obj_id = self.assign_temp_id(obj)
                temp_names.append(obj)
            else:
                obj_id = obj.xml_id
            objs_by_level[ml.name] = obj_id
        for (ml, obj) in objs:
            obj_id = self.get_obj_id(obj)
            ml.get_updown(obj, self, edges)
            objs_dict[obj_id] = (ml, obj)
            if ml.locality in objs_by_level:
                edges.append((objs_by_level[ml.locality], obj_id))
        # 2. topological sort (of keys)
        for k in topsort(edges):
            if k in objs_dict:
                result.append(objs_dict[k])
                del objs_dict[k]
        result += objs_dict.values()
        for obj in temp_names:
            self.clear_temp_id(obj)
        return result

    def register_object(self, obj, schema=None):
        if schema is None:
            mlevel = self.mlevel_for_class(type(obj))
        else:
            mlevel = schema
        if mlevel is None:
            print(self.schema_by_class, file=sys.stderr)
            raise ValueError("No markable level for %s (type %s)" %
                             (obj, type(obj)))
        self.markables_by_start[obj.span[0]].append((mlevel, obj))

    def make_span(self, span):
        wids = self.word_ids
        parts = []
        for start, end in zip(span[::2], span[1::2]):
            if end == start + 1:
                parts.append(wids.get_sym(start))
            else:
                parts.append('%s..%s' % (wids.get_sym(start),
                                         wids.get_sym(end - 1)))
        return ','.join(parts)

    def get_objects_by_class(self, cls, start=0, end=None):
        if end is None:
            end = len(self.words)
        objs_by_start = self.markables_by_start
        mbs_start = objs_by_start.bisect_left(start)
        mbs_end = objs_by_start.bisect_left(end)
        result = []
        for idx in xrange(mbs_start, mbs_end):
            i = objs_by_start.iloc[idx]
            for (mlevel, obj) in objs_by_start[i]:
                if isinstance(obj, cls):
                    result.append(obj)
        return result

    def get_objects_by_level(self, level, start=0, end=None):
        if end is None:
            end = len(self.words)
        objs_by_start = self.markables_by_start
        mbs_start = objs_by_start.bisect_left(start)
        mbs_end = objs_by_start.bisect_left(end)
        result = []
        result = []
        for idx in xrange(mbs_start, mbs_end):
            i = objs_by_start.iloc[idx]
            for (mlevel, obj) in objs_by_start[i]:
                if level == mlevel.name:
                    result.append(obj)
        return result

    def clear_objects_by_level(self, levelname, start=0, end=None):
        if end is None:
            end = len(self.words)
        objs_by_start = self.markables_by_start
        mbs_start = objs_by_start.bisect_left(start)
        mbs_end = objs_by_start.bisect_left(end)
        for idx in xrange(mbs_start, mbs_end):
            i = objs_by_start.iloc[idx]
            objs_new = []
            for (mlevel, obj) in objs_by_start[i]:
                if mlevel.name == levelname:
                    if hasattr(obj, 'xml_id'):
                        del self.object_by_id[obj.xml_id]
                else:
                    objs_new.append((mlevel, obj))
            objs_by_start[i] = objs_new

    def inline_events(self, start, end, levels=None):
        """
        process this part of the document, producing SAX-like events
        
        Note: this interface is expected to change without notice in the 1.0.x releases
        
        :param start: start from here
        :param end: go to this position
        :param levels: if present, only markables on these levels generate events
        """
        objs_by_start = self.markables_by_start
        if end is None:
            end = len(self.words)
        stack = []
        for i, n in izip(xrange(start, end), islice(self.w_objs, start, end)):
            #print("InEv", i, stack)
            # close all tags that must be closed here
            while stack and i >= stack[-1][1]:
                yield ('end', stack[-1][0],)
                stack.pop()
            assert (not stack or stack[-1][1] > i), (i, stack)
            # find all markables starting here
            o_here = objs_by_start.get(i, ())
            #print("InEv pre-filter", o_here)
            if levels is not None:
                o_here = [mlevel_obj for mlevel_obj in o_here if mlevel_obj[0].name in levels]
                #print("InEv post-filter", o_here)
            o_here.sort(key=lambda mlevel_obj: -mlevel_obj[1].span[-1])
            j = 0
            last_o = len(o_here) - 1
            m_here = []
            while j < last_o:
                end_here = o_here[j][1].span[-1]
                if end_here == o_here[j + 1][1].span[-1]:
                    # perform sort by endpoint and topological
                    # sort for coextensive up/down relationships
                    j1 = j + 1
                    while j1 <= last_o and end_here == o_here[j1][1].span[-1]:
                        j1 += 1
                    for mlevel, obj in self.reorder_updown(o_here[j:j1]):
                        m_here.append(mlevel.serialize_object(obj, self))
                    j = j1
                else:
                    mlevel, obj = o_here[j]
                    m_here.append(mlevel.serialize_object(obj, self))
                    j += 1
            while j < len(o_here):
                (mlevel, obj) = o_here[j]
                m_here.append(mlevel.serialize_object(obj, self))
                j += 1
            for m in m_here:
                need_span = False
                endpoint = m[0][-1]
                if len(m[0]) > 2:
                    need_span = True
                    if stack and m[0][-1] > stack[-1][1]:
                        endpoint = stack[-1][1]
                elif stack and m[0][-1] > stack[-1][1]:
                    need_span = True
                    endpoint = stack[-1][1]
                if need_span:
                    m[2]['span'] = self.make_span(m[0])
                yield ('start', m[1], m[2].items(), m[3])
                stack.append((m[1], endpoint))
            yield ('terminal', n)
        # finally, close everything else
        while stack:
            x = stack.pop()
            yield ('end', x[0])

    def write_inline_xml(self, f, start=0, end=None,
                         encoding=None, force_ids=True):
        """inline XML serialization for part or whole of the document"""
        objs_by_start = self.markables_by_start
        if end is None:
            end = len(self.words)
        stack = []
        for i, n in izip(xrange(start, end), islice(self.w_objs, start, end)):
            # close all tags that must be closed here
            while stack and i == stack[-1][1]:
                f.write(' ' * (len(stack) - 1))
                f.write('</%s>\n' % (stack[-1][0],))
                stack.pop()
            assert (not stack or stack[-1][1] > i), (i, stack)
            # find all markables starting here
            o_here = objs_by_start.get(i, ())
            if o_here:
                o_here.sort(key=lambda mlevel_obj: -mlevel_obj[1].span[-1])
                j = 0
                last_o = len(o_here) - 1
                m_here = []
                while j < last_o:
                    end_here = o_here[j][1].span[-1]
                    if end_here == o_here[j + 1][1].span[-1]:
                        # perform sort by endpoint and topological
                        # sort for coextensive up/down relationships
                        j1 = j + 1
                        while j1 <= last_o and end_here == o_here[j1][1].span[-1]:
                            j1 += 1
                        for mlevel, obj in self.reorder_updown(o_here[j:j1]):
                            m_here.append(mlevel.serialize_object(obj, self, force_ids=force_ids))
                        j = j1
                    else:
                        mlevel, obj = o_here[j]
                        m_here.append(mlevel.serialize_object(obj, self, force_ids=force_ids))
                        j += 1
                while j < len(o_here):
                    (mlevel, obj) = o_here[j]
                    m_here.append(mlevel.serialize_object(obj, self, force_ids=force_ids))
                    j += 1
                for m in m_here:
                    need_span = False
                    endpoint = m[0][-1]
                    if len(m[0]) > 2:
                        need_span = True
                        if stack and m[0][-1] > stack[-1][1]:
                            endpoint = stack[-1][1]
                    elif stack and m[0][-1] > stack[-1][1]:
                        need_span = True
                        endpoint = stack[-1][1]
                    if need_span:
                        m[2]['span'] = self.make_span(m[0])
                    open_tag(f, m[1], m[2].items(), len(stack),
                             encoding=encoding)
                    f.write('>\n')
                    for e in m[3]:
                        open_tag(f, e[0], e[1].items(), len(stack) + 1,
                                 encoding=encoding)
                        f.write('/>\n')
                    stack.append((m[1], endpoint))
            t_desc = self.t_schema.serialize_terminal(n, self)
            open_tag(f, t_desc[0], t_desc[1].items(), len(stack),
                     encoding=encoding)
            if t_desc[2]:
                f.write('>\n')
                for e in t_desc[2]:
                    open_tag(f, e[0], e[1].items(), len(stack) + 1,
                             encoding=encoding)
                    f.write('/>\n')
                f.write(' ' * (len(stack)))
                f.write('</%s>\n' % (t_desc[0],))
            else:
                f.write('/>\n')
        # finally, close everything else
        while stack:
            x = stack.pop()
            f.write(' ' * (len(stack) - 1))
            f.write('</%s>\n' % (x[0],))

    def save(self, fname, force_ids=True):
        encoding = 'UTF-8'
        if fname.endswith('.gz'):
            f_out = gzip.open(fname, 'wb')
        else:
            f_out = open(fname, 'wb')
        with f_out:
            print('<?xml version="1.0" encoding="%s"?>' %
                  (encoding,), file=f_out)
            print('<exml-doc>', file=f_out)
            self.describe_schema(f_out, encoding=encoding)
            print('<body serialization="inline">', file=f_out)
            self.write_inline_xml(f_out, encoding=encoding, force_ids=force_ids)
            print('</body>', file=f_out)
            print('</exml-doc>', file=f_out)

    def json_chunk(self, start=0, end=None):
        """
        turns part or whole of the document
        into a JSON fragment
        """
        objs_by_start = self.markables_by_start
        result_by_level = {'_start': start}
        if end is None:
            end = len(self.words)
        terminals = []
        for i in xrange(start, end):
            w_obj = self.w_objs[i]
            terminals.append(self.t_schema.make_json(w_obj, self))
        result_by_level['word'] = terminals
        for i, n in izip(xrange(start, end), islice(self.w_objs, start, end)):
            # find all markables starting here
            o_here = objs_by_start.get(i, ())
            for mlevel, obj in o_here:
                m_levelname = mlevel.name
                m_objs = result_by_level.get(m_levelname)
                if m_objs is None:
                    m_objs = []
                    result_by_level[m_levelname] = m_objs
                m_objs.append(mlevel.make_json(obj, self))
        return result_by_level

    def json_insert(self, json_obj):
        """
        takes the information from a JSON
        fragment and inserts terminals and
        markables into the document
        """
        start = json_obj.get('_start', 0)
        terminals = json_obj['word']
        end = start + len(terminals)
        if start == len(self.words):
            self.w_objs += [self.t_schema.create_from_json(n, self)
                            for n in terminals]
            self.words += [
                w_obj.word for w_obj in self.w_objs[len(self.words):]]
        else:
            assert end <= len(self.words)
            assert self.words[start:end] == [n['form'] for n in terminals]
            self.w_objs[start:end] = [
                self.t_schema.create_from_json(n, self) for n in terminals]
        for i, n in izip(xrange(start, end), islice(self.w_objs, start, end)):
            n.span = [i, i + 1]
            word_id = self.get_obj_id(n)
            assert self.word_ids[word_id] == i
        markables_by_level = {}
        for schema in self.schemas:
            if schema.name not in json_obj:
                continue
            objs = json_obj[schema.name]
            markables = [schema.create_from_json(obj, self) for obj in objs]
            markables_by_level[schema.name] = markables
            for m in markables:
                self.object_by_id[self.get_obj_id(m)] = m
                self.register_object(m)
        for i, n, obj in izip(xrange(start, end), islice(self.w_objs, start, end), terminals):
            self.t_schema.fill_from_json(n, obj, self)
        for schema in self.schemas:
            if schema.name not in json_obj:
                continue
            objs = json_obj[schema.name]
            for obj, n in izip(json_obj[schema.name], markables_by_level[schema.name]):
                schema.fill_from_json(n, obj, self)

    def clear_markables(self, start=0, end=None):
        if end is None:
            if start == 0:
                self.markables_by_start.clear()
                self.w_objs = [None]*len(self.w_objs)
                return
            end = len(self.words)
        mbs = self.markables_by_start
        mbs_start = mbs.bisect_left(start)
        mbs_end = mbs.bisect_left(end)
        del mbs.iloc[mbs_start:mbs_end]
        for i in xrange(start, end):
            self.w_objs[i] = None

    def describe_schema(self, f, encoding=None):
        edge_descrs = {}
        f.write("<schema>\n")
        self.t_schema.describe_schema(f, edge_descrs, encoding=encoding)
        for schema in self.schemas:
            schema.describe_schema(f, edge_descrs, encoding=encoding)
        for (name, (schema, parents)) in edge_descrs.items():
            open_tag(f, "edge",
                     [('name', name), ('parent', '|'.join(parents))],
                     encoding=encoding)
            f.write('>\n')
            for att in schema.attributes:
                att.describe_schema(f, encoding=encoding)
            f.write('</edge>\n')
        f.write("</schema>\n")


def assign_node_ids(n, prefix, sent_start=0):
    n.span = [n.start + sent_start, n.end + sent_start]
    if hasattr(n, 'xml_id'):
        pass
    elif hasattr(n, 'id'):
        n.xml_id = '%s_%s' % (prefix, n.id)
    for n1 in n.children:
        assign_node_ids(n1, prefix, sent_start)


class GenericMarkable(object):

    def __init__(self, **kw):
        self.__dict__.update(kw)


class Text(object):

    def __init__(self, origin, doc_no=0):
        self.origin = origin
        self.doc_no = doc_no


class NamedEntity(object):

    def __init__(self, kind, **kw):
        self.kind = kind
        self.__dict__.update(kw)


rel_map = {
    'evaluative': 'antithesis',
    'epistemic_cause': 'evidence'
}


def get_rel(info, which):
    k = getattr(info, which, None)
    if k in rel_map:
        k = rel_map[k]
    if k == 'NULL':
        k = None
    return k


def make_syntax_doc(want_ne=True, want_wsd=False,
                    want_deps=False):
    """
    Creates a :py:class:`Document` object with a TeuBa-D/Z compatible
    markable scheme.
    """
    text_schema = MarkableSchema('text', Text)
    text_schema.attributes = [TextAttribute('origin')]
    text_schema.init_attrs = text_schema.attributes
    s_schema = MarkableSchema('sentence', tree.Tree)
    s_schema.locality = 'text'
    nt_schema = MarkableSchema('node', tree.NontermNode)
    nt_schema.locality = 'sentence'
    secedge_edge = SecondaryEdges('secEdge')
    relation_edge = ReferenceEdges('relation')
    split_edge = SplitRefEdges('splitRelation')
    nt_schema.attributes = [EnumAttribute('cat'),
                            EnumAttribute('func', prop_name='edge_label'),
                            RefAttribute('parent', restriction='up',
                                         restrict_target=['node']),
                            TextAttribute('comment')]
    nt_schema.init_attrs = nt_schema.attributes[:1]
    nt_schema.edges = [secedge_edge,
                       relation_edge,
                       split_edge]
    ne_schema = MarkableSchema('ne', NamedEntity)
    ne_schema.locality = 'sentence'
    ne_schema.attributes = [EnumAttribute('type', prop_name='kind')]
    ne_schema.init_attrs = ne_schema.attributes
    t_schema = TerminalSchema('word', tree.TerminalNode)
    t_schema.attributes = [TextAttribute('form', prop_name='word'),
                           EnumAttribute('pos', prop_name='cat'),
                           EnumAttribute('morph', default_val='--'),
                           TextAttribute('lemma', default_val='--'),
                           EnumAttribute('func', prop_name='edge_label'),
                           RefAttribute('parent', restriction='up',
                                        restrict_target=['node']),
                           TextAttribute('comment')]
    t_schema.edges = [secedge_edge,
                      relation_edge,
                      split_edge]
    all_schemas = [s_schema, nt_schema, text_schema]
    if want_ne:
        all_schemas.append(ne_schema)
    if want_wsd:
        t_schema.attributes[-1:-1] = [TextAttribute(
            'wsd-lexunits', prop_name='wsd_lexunits'),
            TextAttribute('wsd-comment', prop_name='wsd_comment')]
    if want_deps:
        t_schema.attributes[-1:-1] = [RefAttribute(
            'dephead', prop_name='syn_parent', restrict_target=['word']),
            EnumAttribute('deprel', prop_name='syn_label')]
    return Document(t_schema, all_schemas)


def make_noderef(x):
    sentid, nodeid = x.split(':')
    try:
        nid = int(nodeid)
    except ValueError:
        return 's%s_%s' % (sentid, nodeid)
    else:
        if nid < 500:
            nid += 1
        return 's%s_%d' % (sentid, nid)


def make_parts(xs, offset=0):
    spans = []
    current = None
    for x in xs:
        if current:
            if current[1] == x:
                current[1] = x + 1
            else:
                spans.append(current[0] + offset)
                spans.append(current[1] + offset)
                current = [x, x + 1]
        else:
            current = [x, x + 1]
    if current:
        spans.append(current[0] + offset)
        spans.append(current[1] + offset)
    return spans


def add_tree_to_doc(t, ctx, start=None):
    """
    adds a :py:class:`pytree.tree.Tree` instance
    to the document, extending terminals and the
    node and sentence markable levels
    """
    if start is None:
        sent_start = len(ctx.words)
    else:
        sent_start = start
    if hasattr(t, 'sent_no'):
        prefix = 's%s' % (t.sent_no,)
        t.xml_id = prefix
    elif hasattr(t, 'xml_id'):
        prefix = t.xml_id
    else:
        assert False, t
    for i, n in enumerate(t.terminals):
        n.xml_id = '%s_%d' % (prefix, i + 1)
    for n in t.roots:
        assign_node_ids(n, prefix, sent_start)
    if hasattr(t, 'all_nes'):
        last_num = defaultdict(int)
        suffixes = ['', 'a', 'b', 'c', 'd']
        for kind, local_span in t.all_nes:
            ne = NamedEntity(kind)
            assert local_span, ('Empty NE span in sentence %s' % (t.sent_no,))
            ne.span = [k + sent_start for k in local_span]
            ne_start = ne.span[0]
            suff = suffixes[last_num[ne_start]]
            last_num[ne_start] += 1
            ne.xml_id = 'ne_%s%s' % (ne_start, suff)
            ctx.register_object(ne)
    if start is None:
        for n in t.terminals:
            ctx.add_terminal(n)
    else:
        for i, n in enumerate(t.terminals):
            ctx.replace_terminal(start + i, n)
    t.span = [sent_start, sent_start + len(t.terminals)]
    ctx.register_object(t)
    for n in t.node_table.values():
        ctx.register_object(n)


def postprocess_doc(doc, start=0, end=None):
    for t in doc.get_objects_by_class(tree.Tree, start, end):
        t.terminals = doc.w_objs[t.span[0]:t.span[-1]]
        all_nonterm = doc.get_objects_by_class(
            tree.NontermNode, t.span[0], t.span[-1])
        for n in all_nonterm:
            n.children = []
            n.start = n.span[0] - t.span[0]
            n.end = n.span[-1] - t.span[0]
        roots = []
        for i, n in enumerate(t.terminals):
            if n.parent is not None:
                n.parent.children.append(n)
            else:
                roots.append(n)
            n.start = i
            n.end = i + 1
        for n in all_nonterm:
            if n.parent is not None:
                n.parent.children.append(n)
            else:
                roots.append(n)
        roots.sort(key=lambda n: n.span[0])
        t.roots = roots
        for n in all_nonterm:
            assert n.children, (n.span, n.cat, n.xml_id, t.span, t.xml_id)
            n.children.sort(key=lambda n1: n1.span[0])


def read_json_doc(f_in, make_schema=None):
    """
    reads a JSON file with a single document
    according to the syntax_doc(TüBa-D/Z) schema
    """
    if make_schema is None:
        make_schema = make_syntax_doc
    doc = make_schema()
    json_obj = json.load(f_in)
    doc.json_insert(json_obj)
    postprocess_doc(doc)
    return doc

def normalize_encoding(encoding):
    """
    This function tries to return  
    :param encoding: 
    :return: 
    """
    if encoding is None:
        return None
    if encoding.upper().replace('-', '') == 'UTF8':
        return 'UTF-8'
    if encoding.upper().replace('-', '') in ['ISO88591', 'ISO885915', 'LATIN1', 'CP850']:
        return 'iso8859-1'
    else:
        if encoding.upper().startswith('ISO8859'):
            return 'ISO-8859'+encoding[7:]
        elif encoding.upper().startswith('ISO-'):
            return encoding.upper()
        else:
            return encoding.lower()

class XMLCorpusReader(object):
    """
    Code to read an EXML corpus in XML form.
    This is currently considered experimental.
    """

    def __init__(self, doc, fname, encoding='UTF-8'):
        """
        
        :param doc: a exmldoc.Document
        :type doc: Document
        :param fname: a filename
        :param encoding: the encoding for the corpus
        """
        self.doc = doc
        self.fname = fname
        if fname.endswith('.gz'):
            f_in = gzip.open(fname, 'rb')
        else:
            f_in = open(fname, 'rb')
        if have_lxml:
            # try to recover from XML problems
            self.parse = etree.iterparse(f_in, events=('start', 'end',), recover=True)
        else:
            self.parse = etree.iterparse(f_in, events=('start', 'end',))
        self.state = 'BEFORE_HEAD'
        self.encoding = normalize_encoding(encoding)
        self.markable_stack = []
        self.old_posn = 0
        self.at_end = False

    def read_header(self):
        # read until end of header
        if self.state != 'BEFORE_HEAD':
            assert False
        while True:
            evt, elem = next(self.parse)
            if evt == 'end' and elem.tag == 'schema':
                process_schema(self.doc, elem)
                self.state = 'BEFORE_BODY'
                return

    def addNext(self):
        # read until end-of-text or end-of-body and return last_stop
        if self.state in ['BEFORE_HEAD']:
            self.read_header()
        doc = self.doc
        markable_stack = self.markable_stack
        encoding = self.encoding
        last_stop = len(doc.words)
        cur_pos = last_stop
        temp_ids = []
        while self.state == 'BEFORE_BODY':
            evt, elem = next(self.parse)
            if evt == 'start' and elem.tag == 'body':
                self.state = 'IN_BODY'
        if self.state == 'AT_END':
            self.at_end = True
            raise StopIteration()
        while True:
            try:
                evt, elem = next(self.parse)
            except StopIteration:
                self.state = 'AT_END'
                return len(doc.words)
            if evt == 'end' and elem.tag == 'body':
                # wrap up any loose ends
                for chld in elem.getchildren():
                    fill_attributes(chld, doc, encoding)
                if last_stop != self.old_posn:
                    self.old_posn = last_stop
                    return last_stop
            elif evt == 'start':
                # create markable
                # if a markable or word does not have an XML-id,
                # assign one by default
                if elem.tag == 'word':
                    schema = doc.t_schema
                    obj = schema.create_from_xml(elem, doc, encoding)
                    obj.span = [cur_pos, None]
                    #print >>sys.stderr, "found word: %s"%(obj.word,)
                    doc.add_terminal(obj)
                    doc.object_by_id[obj.xml_id] = obj
                    in_word = True
                else:
                    # set start point
                    try:
                        schema = doc.schema_by_name(elem.tag)
                    except KeyError:
                        # assume it's an edge; we'll deal with it later
                        if in_word:
                            (schema, obj) = (doc.t_schema, doc.words[-1])
                        elif markable_stack:
                            (schema, obj) = markable_stack[-1]
                        else:
                            schema = obj = None
                        try:
                            edge_schema = schema.edge_by_name(elem.tag)
                        except KeyError:
                            edge_schema = None
                        if edge_schema is not None:
                            print("edge: %s" % (elem.tag), file=sys.stderr)
                        else:
                            print("No schema:", elem.tag, [
                                  s.name for s in doc.schemas])
                    else:
                        obj = schema.create_from_xml(elem, doc, encoding)
                        obj.span = [cur_pos, None]
                        markable_stack.append((schema, obj))
            elif evt == 'end':
                # print elem.tag, markable_stack
                if elem.tag == 'word':
                    in_word = False
                    cur_pos += 1
                elif elem.tag == 'body':
                    for chld in elem.getchildren():
                        fill_attributes(chld, doc, encoding)
                    elem.clear()
                    # TODO wrap up stuff
                    self.old_posn = doc.size()
                    return doc.size()
                elif markable_stack and elem.tag == markable_stack[-1][0].name:
                    # set end point of markable
                    (schema, obj) = markable_stack.pop()
                    obj.span[1] = cur_pos
                    doc.register_object(obj, schema)
                if elem.tag in ['text', 'doc']:
                    fill_attributes(elem, doc, encoding)
                    elem.clear()
                    self.old_posn = last_stop
                    return last_stop

    def inline_events(self, levels=None, clean=True):
        """
        yields a sequence of events for a corpus, reading in the corpus
        up to each text boundary and resolving references.
        
        Note: this interface is expected to change without notice in the 1.0.x releases
        
        :return: a sequence of events 
        """
        last_stop = len(self.doc.words)
        while True:
            try:
                new_stop = self.addNext()
                #print("InEv from:", last_stop, "to:", new_stop)
                for ev in self.doc.inline_events(last_stop, new_stop, levels):
                    yield ev
                if clean:
                    self.doc.clear_markables(last_stop, new_stop)
                last_stop = new_stop
            except StopIteration:
                #print("InEvStop from:", last_stop, "to:", new_stop)
                for ev in self.doc.inline_events(last_stop, len(self.doc.words), levels):
                    yield ev
                break

class JSONCorpusReader:

    def __init__(self, doc, fname):
        """
        initializes the corpus reader

        :param doc: a Document that we will read into
        :type doc Document:
        :param fname: the file name of the document to be read
        """
        self.doc = doc
        self.fname = fname
        self.f = open(fname, 'rb')

    def addNext(self):
        # TODO: basic consistency check to ensure
        # that JSON offsets are right
        l = self.f.readline()
        if l == '':
            raise StopIteration
        last_stop = len(self.doc.words)
        obj = json.loads(l)
        self.doc.json_insert(obj)
        return last_stop


def read_trees_exml(fname):
    """
    reads in an EXML file and returns a sequence of trees,
    similar to read_trees() does for treebank files
    """
    # TODO add a parameter for reading in JSON-format EXML files?
    doc = make_syntax_doc(want_deps=True)
    reader = XMLCorpusReader(doc, fname)
    last_stop = len(doc.words)
    while True:
        try:
            new_stop = reader.addNext()
            print("last_stop=%d new_stop=%d" % (
                last_stop, new_stop), file=sys.stderr)
            if new_stop != last_stop:
                trees = doc.get_objects_by_class(
                    tree.Tree, last_stop, new_stop)
                print("%d sentences in %d tokens" % (
                    len(trees), new_stop - last_stop), file=sys.stderr)
                for t in trees:
                    # TODO fill terminals
                    t.terminals = doc.w_objs[t.span[0]:t.span[-1]]
                    for i, n in enumerate(t.terminals):
                        n.start = i
                        n.end = i + 1
                    # TODO fill roots
                    # TODO set start/end of NT nodes
                    yield t
                last_stop = new_stop
        except StopIteration:
            trees = doc.get_objects_by_class(
                tree.Tree, last_stop, len(doc.words))
            assert len(trees) == 0
            break

def create_doc(extra_word_attrs=None, extra_levels=None, **extra):
    doc = make_syntax_doc(want_deps=True)
    if extra_word_attrs is not None:
        t_schema = doc.t_schema
        for att in extra_word_attrs:
            t_schema.add_attribute(att)
    if extra_levels is not None:
        doc.add_schemas(extra_levels)
    for level in doc.schemas:
        key = "extra_"+level.name
        if key in extra:
            extra_attrs = extra[key]
            for att in extra_attrs:
                level.add_attribute(att)
    return doc

def load(fname, extra_word_attrs=None, extra_levels=None, encoding=None, **extra):
    """
    reads an EXML document as produced by ExmlPipe

    :param fname: the filename of the EXML document 
    :return: an exmldoc.Document
    """
    doc = create_doc(extra_word_attrs, extra_levels, **extra)
    reader = XMLCorpusReader(doc, fname, encoding)
    last_stop = len(doc.words)
    while True:
        try:
            reader.addNext()
        except StopIteration:
            postprocess_doc(doc)
            return doc


def write_corpus_xml(doc, reader, f_out, encoding="ISO-8859-15"):
    """
    writes a corpus as an ExportXMLv2 xml file
    """
    print('<?xml version="1.0" encoding="%s"?>' % (encoding,), file=f_out)
    print('<exml-doc>', file=f_out)
    doc.describe_schema(f_out)
    # do the actual conversion
    print('<body serialization="inline">', file=f_out)
    last_stop = len(doc.words)
    while True:
        try:
            new_stop = reader.addNext()
            if (new_stop != last_stop):
                doc.write_inline_xml(f_out, last_stop, new_stop)
                doc.clear_markables(last_stop, new_stop)
                last_stop = new_stop
        except StopIteration:
            break
    doc.write_inline_xml(f_out, last_stop)
    print('</body>', file=f_out)
    print('</exml-doc>', file=f_out)


def write_corpus_json(doc, reader, f_out):
    """
    writes a corpus as one JSON expression per document
    """
    last_stop = len(doc.words)
    while True:
        try:
            new_stop = reader.addNext()
            if (new_stop != last_stop):
                print(json.dumps(doc.json_chunk(last_stop, new_stop)), file=f_out)
                doc.clear_markables(last_stop, new_stop)
                last_stop = new_stop
        except StopIteration:
            break
    print(json.dumps(doc.json_chunk(last_stop)), file=f_out)


def process_node_schema(doc, schema, elem):
    """
    processes the part regarding one markable in the EXML document
    schema
    """
    for chld in elem.getchildren():
        if chld.tag == 'text-attr':
            tag_name = chld.attrib['name']
            try:
                att = schema.attribute_by_name(tag_name)
            except KeyError:
                print("Undeclared text attribute: %s.%s" % (
                    schema.name, tag_name,), file=sys.stderr)
                att = TextAttribute(tag_name, '_auto_%s' % (tag_name,))
                schema.add_attribute(att)
        elif chld.tag == 'enum-attr':
            tag_name = chld.attrib['name']
            try:
                att = schema.attribute_by_name(tag_name)
            except KeyError:
                print("Undeclared enum attribute: %s.%s" % (
                    schema.name, tag_name,), file=sys.stderr)
                att = EnumAttribute(tag_name,
                                    prop_name='_auto_%s' % (tag_name,))
                schema.add_attribute(att)
            for elm_val in chld.findall('val'):
                # TODO check what happens with double attributes
                att.add_item(elm_val.attrib['name'],
                             elm_val.attrib.get('description', None))
        elif chld.tag == 'node-ref':
            tag_name = chld.attrib['name']
            try:
                att = schema.attribute_by_name(tag_name)
            except KeyError:
                print("Undeclared node reference: %s.%s" % (
                    schema.name, tag_name,), file=sys.stderr)
                att = RefAttribute(tag_name,
                                   prop_name='_auto_%s' % (tag_name,))
                schema.add_attribute(att)
        else:
            print('Unkown tag in node schema: %s' %
                  (chld.tag,), file=sys.stderr)


def process_schema(doc, elem):
    """given an EXML document doc and a XML element elem
    with the attributes, add stuff to the schema
    """
    for chld in elem.getchildren():
        if chld.tag == 'tnode':
            node_name = chld.attrib['name']
            assert node_name == 'word'
            process_node_schema(doc, doc.t_schema, chld)
        elif chld.tag == 'node':
            node_name = chld.attrib['name']
            try:
                schema = doc.schema_by_name(node_name)
            except KeyError:
                print("Undeclared markable level: %s" %
                      (node_name,), file=sys.stderr)
                # TODO add locality information
                cls = type('auto_markable_%s' % (node_name,),
                           (GenericMarkable,), {})
                schema = MarkableSchema(node_name, cls)
                doc.add_schemas([schema])
            process_node_schema(doc, schema, chld)
        elif chld.tag == 'edge':
            edge_name = chld.attrib['name']
            node_names = chld.attrib['parent'].split('|')
            for node_name in node_names:
                if node_name == 'word':
                    schema = doc.t_schema
                else:
                    schema = doc.schema_by_name(node_name)
                try:
                    edge_schema = schema.edge_by_name(edge_name)
                except KeyError:
                    # TODO add to markable level
                    print("Undeclared edge schema: %s.%s" % (
                        node_name, edge_name), file=sys.stderr)
                    edge_schema = GenericEdges(edge_name,
                                               prop_name='auto_%s' % (edge_name,))
                    schema.add_edge(edge_schema)
                process_node_schema(doc, edge_schema, chld)


def fill_attributes(elem, doc, encoding=None):
    try:
        xml_id = elem.attrib[QNAME_XML_ID]
        obj = doc.object_by_id[xml_id]
        doc.clear_temp_id(obj)
    except KeyError:
        if elem.tag in ['text', 'doc'] and not elem.getchildren():
            # Presumably already cleared, nothing to see
            pass
        else:
            print("No ID: %s" % (elem,), file=sys.stderr)
        return
    if elem.tag == 'word':
        schema = doc.t_schema
    else:
        schema = doc.schema_by_name(elem.tag)
    schema.fill_from_xml(obj, elem, doc, encoding)
    for chld in elem.getchildren():
        if chld.tag == 'word':
            fill_attributes(chld, doc, encoding)
        else:
            try:
                c_schema = doc.schema_by_name(chld.tag)
            except KeyError:
                e_schema = schema.edge_by_name(chld.tag)
                edges = e_schema.get_edges(obj, doc)
                val = []
                for att in e_schema.attributes:
                    if att.name in chld.attrib:
                        val.append(att.unmap_attr(chld.attrib[att.name],
                                                  doc, encoding))
                    else:
                        print("No val for %s.%s.%s" % (
                            schema.name, e_schema.name, att.name))
                        val.append(None)
                # print e_schema.name, val
                edges.append(val)
                e_schema.set_edges(obj, edges, doc)
                continue
            else:
                fill_attributes(chld, doc, encoding)



from __future__ import print_function
import optparse
import xml.etree.cElementTree as etree
from exmldoc import make_syntax_doc, process_schema, fill_attributes


def exml_lint_main(argv=None):
    """
    Reads an EXML file and writes it back in normal form,
    allowing a conversion between ExportXML and EXML-JSON
    
    This file also demonstrates the use of exmldoc to stream
    a larger document
    """
    encoding = 'UTF-8'
    oparse = optparse.OptionParser()
    opts, args = oparse.parse_args(argv)
    ctx = etree.iterparse(args[0], events=('start', 'end'))
    doc = make_syntax_doc()
    f_out = open(args[1], 'wb')
    state = 'BEFORE_HEAD'
    markable_stack = []
    cur_pos = 0
    last_stop = 0
    in_word = False
    for evt, elem in ctx:
        if state == 'BEFORE_HEAD' and evt == 'end' and elem.tag == 'schema':
            process_schema(doc, elem)
            print('<?xml version="1.0" encoding="%s"?>' %
                  (encoding,), file=f_out)
            print('<exml-doc>', file=f_out)
            doc.describe_schema(f_out, encoding=encoding)
            state = 'BEFORE_BODY'
        elif state == 'BEFORE_BODY' and evt == 'start' and elem.tag == 'body':
            state = 'IN_BODY'
            print('<body serialization="inline">', file=f_out)
        elif state == 'IN_BODY':
            if evt == 'end' and elem.tag == 'body':
                # TODO wrap up all loose ends
                pass
            elif evt == 'start':
                # create markable
                # TODO if a markable or word does not have an XML-id,
                # assign one by default
                if elem.tag == 'word':
                    schema = doc.t_schema
                    obj = schema.create_from_xml(elem, doc, encoding)
                    obj.span = [cur_pos, None]
                    doc.add_terminal(obj)
                    doc.object_by_id[obj.xml_id] = obj
                    in_word = True
                else:
                    # set start point
                    try:
                        schema = doc.schema_by_name(elem.tag)
                    except KeyError:
                        # assume it's an edge; we'll deal with it later
                        if in_word:
                            (schema, obj) = (doc.t_schema, doc.words[-1])
                        elif markable_stack:
                            (schema, obj) = markable_stack[-1]
                        else:
                            schema = obj = None
                        try:
                            edge_schema = schema.edge_by_name(elem.tag)
                        except KeyError:
                            edge_schema = None
                        if edge_schema is not None:
                            pass
                        else:
                            print("No schema:", elem.tag, [
                                  s.name for s in doc.schemas])
                    else:
                        obj = schema.create_from_xml(elem, doc, encoding)
                        obj.span = [cur_pos, None]
                        markable_stack.append((schema, obj))
            elif evt == 'end':
                # print elem.tag, markable_stack
                if elem.tag == 'word':
                    in_word = False
                    cur_pos += 1
                elif elem.tag == 'body':
                    fill_attributes(elem, doc, encoding)
                    elem.clear()
                    # write out part of the document
                    doc.write_inline_xml(f_out, last_stop, cur_pos)
                    doc.clear_markables(last_stop, cur_pos)
                    last_stop = cur_pos
                    state = 'AFTER_BODY'
                elif markable_stack and elem.tag == markable_stack[-1][0].name:
                    # set end point of markable
                    (schema, obj) = markable_stack.pop()
                    obj.span[1] = cur_pos
                    doc.register_object(obj, schema)
                if elem.tag == 'text':
                    # if it's a text markable, fill attributes and empty out
                    # the element
                    fill_attributes(elem, doc, encoding)
                    elem.clear()
                    # write out part of the document
                    doc.write_inline_xml(f_out, last_stop, cur_pos)
                    doc.clear_markables(last_stop, cur_pos)
                    last_stop = cur_pos
    print('</body>', file=f_out)
    print('</exml-doc>', file=f_out)


if __name__ == '__main__':
    exml_lint_main()



class PythonAlphabet(object):
    def __init__(self):
        self.int2obj = []
        self.obj2int = {}

    def __getitem__(self, k):
        if k in self.obj2int:
            return self.obj2int[k]
        else:
            n = len(self.int2obj)
            self.int2obj.append(k)
            self.obj2int[k] = n
            return n

    def __len__(self):
        return len(self.int2obj)

    def __iter__(self):
        return iter(self.int2obj)

    @property
    def words(self):
        return self.int2obj



import unittest
from exmldoc.alphabet import PythonAlphabet

class TestAlphabet(unittest.TestCase):
    def test_alphabet_basic(self):
        alph = PythonAlphabet()
        idx1 = alph[1]
        idx2 = alph['a']
        idx3 = alph[(1,2)]
        idx4 = alph['a']
        self.assertEqual(idx1, 0,
                         'indices must start at 0')
        self.assertEqual(idx2, idx4,
                         'the same object must always get the same index')
        self.assertEqual(idx3 + 1, len(alph),
                         'length should be last index plus one')
        self.assertEqual(list(alph), [1, 'a', (1, 2)],
                         'iteration should return objects in the correct order')


from setuptools import setup, find_packages

setup(
    name="exmldoc",
    version="1.0.6",
    author='Yannick Versley',
    author_email='yversley@gmail.com',
    description='Deals with ExportXMLv2 format for annotated text',
    install_requires=[
        'isounidecode',
        'simplejson',
        'msgpack-python',
        'sortedcontainers',
        'setuptools >= 19',
        'mock >= 2.0.0'
    ],
    license='LGPLv3',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Text Processing :: Linguistic'],
    package_dir = {'': 'py_src'},
    packages=find_packages('py_src'),
    test_suite = 'exmldoc.tests',
    entry_points={
        'console_scripts': [
            'exml2cqp = exmldoc.exml2cqp:main'
        ]
    }
)
